from .random_client import roundtrip
from fix_plugin_gcp.resources.storage import *
from fix_plugin_gcp.resources.base import GraphBuilder


def test_gcp_billing_account(random_builder: GraphBuilder) -> None:
    roundtrip(GcpBucket, random_builder)



import json
import os
from pathlib import Path
from queue import Queue
from typing import List

import yaml

from fix_plugin_gcp import GcpConfig
from fix_plugin_gcp.collector import GcpProjectCollector, all_resources, called_collect_apis, called_mutator_apis
from fix_plugin_gcp.gcp_client import GcpApiSpec
from fix_plugin_gcp.resources.base import GcpProject, GraphBuilder
from fix_plugin_gcp.resources.billing import GcpSku
from fix_plugin_gcp.resources.compute import GcpMachineType
from fixlib.baseresources import Cloud
from fixlib.config import current_config
from fixlib.core.actions import CoreFeedback
from fixlib.graph import Graph


def collector_with_graph(graph: Graph) -> GcpProjectCollector:
    collector = GcpProjectCollector(
        config=GcpConfig(),
        cloud=Cloud(id="gcp"),
        project=GcpProject(id="test"),
        core_feedback=CoreFeedback("test", "test", "test", Queue()),
    )
    collector.graph = graph
    return collector


def test_project_collection(random_builder: GraphBuilder) -> None:
    # create the collector from the builder values
    config: GcpConfig = current_config().gcp
    project = GcpProjectCollector(config, random_builder.cloud, random_builder.project, random_builder.core_feedback)
    # use the graph provided by the random builder - it already has regions and zones
    # the random builder will not create new regions or zones during the test
    project.graph = random_builder.graph
    project.collect()
    # the number of resources in the graph is not fixed, but it should be at least the number of resource kinds
    assert len(project.graph.nodes) >= len(all_resources)


def test_remove_unconnected_nodes(random_builder: GraphBuilder) -> None:
    with open(os.path.dirname(__file__) + "/files/machine_type.json") as f:
        GcpMachineType.collect(raw=json.load(f)["items"]["machineTypes"], builder=random_builder)
    with open(os.path.dirname(__file__) + "/files/skus.json") as f:
        GcpSku.collect(raw=json.load(f)["skus"], builder=random_builder)

    collector = collector_with_graph(random_builder.graph)

    num_all_machine_types = len(list(collector.graph.search("kind", "gcp_machine_type")))
    num_all_skus = len(list(collector.graph.search("kind", "gcp_sku")))

    collector.remove_unconnected_nodes()

    assert len(list(collector.graph.search("kind", "gcp_machine_type"))) < num_all_machine_types
    assert len(list(collector.graph.search("kind", "gcp_sku"))) < num_all_skus


def test_role_creation() -> None:
    def iam_role_for(name: str, description: str, calls: List[GcpApiSpec], file: bool = False) -> str:
        permissions = sorted({p for api in calls for p in api.iam_permissions})
        result = yaml.safe_dump(
            {"title": name, "description": description, "stage": "GA", "includedPermissions": permissions},
            sort_keys=False,
        )
        if file:
            with open(Path.home() / f"{name}.yaml", "w") as f:
                f.write(result)
        return result

    write_files = False
    c = iam_role_for("fix_access", "Permissions required to collect resources.", called_collect_apis(), write_files)
    m = iam_role_for("fix_mutate", "Permissions required to mutate resources.", called_mutator_apis(), write_files)
    assert c is not None
    assert m is not None



from .random_client import FixturedClient, roundtrip, json_roundtrip
from fix_plugin_gcp.resources.billing import *
from fix_plugin_gcp.resources.base import GraphBuilder


def test_gcp_billing_account(random_builder: GraphBuilder) -> None:
    roundtrip(GcpBillingAccount, random_builder)
    json_roundtrip(GcpProjectBillingInfo, random_builder)
    assert len(random_builder.edges_of(GcpBillingAccount, GcpProjectBillingInfo)) > 0


def test_gcp_service(random_builder: GraphBuilder) -> None:
    roundtrip(GcpService, random_builder)


def test_gcp_service_and_sku(random_builder: GraphBuilder) -> None:
    SERVICE_ID = "services/6F81-5844-456A"
    SERVICE_NAME = "Compute Engine"
    fixture_replies = {
        "Service": {"serviceId": lambda: SERVICE_ID, "displayName": lambda: SERVICE_NAME},
        "Sku": {"name": lambda: SERVICE_ID},
    }
    with FixturedClient(random_builder, fixture_replies) as random_builder:
        services = GcpService.collect_resources(random_builder)
        random_builder.executor.wait_for_submitted_work()
        for service in services:
            service.connect_in_graph(random_builder, {"Dummy": "Source"})
        assert len(random_builder.edges_of(GcpService, GcpSku)) > 0



import multiprocessing
from concurrent import futures
from typing import Optional, Dict, Any

import fixlib.proc
from fixlib.args import ArgumentParser
from fixlib.args import Namespace
from fixlib.baseplugin import BaseCollectorPlugin
from fixlib.baseresources import Cloud
from fixlib.config import Config, RunningConfig
from fixlib.core.actions import CoreFeedback
from fixlib.graph import Graph
from fixlib.logger import log, setup_logger
from .collector import GcpProjectCollector
from .config import GcpConfig
from .resources.base import GcpProject
from .utils import Credentials


class GCPCollectorPlugin(BaseCollectorPlugin):
    """Google Cloud Platform fix collector plugin.

    Gets instantiated in fix's Processor thread. The collect() method
    is run during a resource collection loop.
    """

    cloud = "gcp"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.core_feedback: Optional[CoreFeedback] = None

    def collect(self) -> None:
        """Run by fix during the global collect() run.

        This method kicks off code that adds GCP resources to `self.graph`.
        When collect() finishes the parent thread will take `self.graph` and merge
        it with the global production graph.
        """
        log.debug("plugin: GCP collecting resources")
        assert self.core_feedback, "core_feedback is not set"  # will be set by the outer collector plugin
        feedback = self.core_feedback.with_context("gcp")

        cloud = Cloud(id=self.cloud, name="Gcp")

        credentials = Credentials.all(feedback)
        log.debug(f"Found {len(credentials)} GCP projects total")
        if len(Config.gcp.project) > 0:
            for project in list(credentials.keys()):
                if project not in Config.gcp.project:
                    log.debug(f"Skipping project {project} because it is not in the configured projects list")
                    del credentials[project]

        if len(credentials) == 0:
            return

        max_workers = (
            len(credentials) if len(credentials) < Config.gcp.project_pool_size else Config.gcp.project_pool_size
        )
        collect_args = {}
        pool_args = {"max_workers": max_workers}
        pool_executor = futures.ThreadPoolExecutor
        if Config.gcp.fork_process:
            collect_args = {
                "args": ArgumentParser.args,
                "running_config": Config.running_config,
                "credentials": credentials if all(v is None for v in credentials.values()) else None,
            }
            collect_method = collect_in_process
        else:
            collect_method = self.collect_project

        with pool_executor(**pool_args) as executor:
            # noinspection PyTypeChecker
            wait_for = [
                executor.submit(collect_method, project_id, feedback, cloud, **collect_args)
                for project_id in credentials.keys()
            ]
            for future in futures.as_completed(wait_for):
                project_graph = future.result()
                if not isinstance(project_graph, Graph):
                    log.error(f"Skipping invalid project_graph {type(project_graph)}")
                    continue
                self.send_account_graph(project_graph)
                del project_graph

    @staticmethod
    def collect_project(
        project_id: str,
        core_feedback: CoreFeedback,
        cloud: Cloud,
        args: Optional[Namespace] = None,
        running_config: Optional[RunningConfig] = None,
        credentials: Optional[Dict[str, Any]] = None,
    ) -> Optional[Graph]:
        """Collects an individual project.

        Is being called in collect() and either run within a thread or a spawned
        process. Depending on whether `gcp.fork_process` was specified or not.

        Because the spawned process does not inherit any of our memory or file
        descriptors we are passing the already parsed `args` Namespace() to this
        method.
        """
        project = GcpProject(id=project_id, name=project_id)
        collector_name = f"gcp_{project_id}"
        fixlib.proc.set_thread_name(collector_name)

        if args is not None:
            ArgumentParser.args = args
            setup_logger("fixworker-gcp", force=True, level=getattr(args, "log_level", None))
        if running_config is not None:
            Config.running_config.apply(running_config)

        if credentials is not None:
            Credentials._credentials = credentials
            Credentials._initialized = True

        log.debug(f"Starting new collect process for project {project.dname}")

        try:
            core_feedback.progress_done(project_id, 0, 1)
            gpc = GcpProjectCollector(Config.gcp, cloud, project, core_feedback)
            gpc.collect()
            core_feedback.progress_done(project_id, 1, 1)
        except Exception as ex:
            core_feedback.with_context("gcp", project_id).error(f"Failed to collect project: {ex}", log)
            return None
        else:
            return gpc.graph

    @staticmethod
    def add_config(config: Config) -> None:
        """Called by fix upon startup to populate the Config store"""
        config.add_config(GcpConfig)


def collect_project_proxy(*args, queue: multiprocessing.Queue, **kwargs) -> None:  # type: ignore
    fixlib.proc.collector_initializer()
    queue.put(GCPCollectorPlugin.collect_project(*args, **kwargs))


def collect_in_process(*args, **kwargs) -> Optional[Graph]:  # type: ignore
    ctx = multiprocessing.get_context("spawn")
    queue = ctx.Queue()
    kwargs["queue"] = queue
    process = ctx.Process(target=collect_project_proxy, args=args, kwargs=kwargs)
    process.start()
    graph = queue.get()
    process.join()
    return graph  # type: ignore



import json
import os
from fix_plugin_gcp.resources.compute import *
from fix_plugin_gcp.resources.billing import GcpSku
from .random_client import roundtrip, connect_resource, FixturedClient
from fix_plugin_gcp.resources.base import GraphBuilder, GcpRegion


def test_gcp_accelerator_type(random_builder: GraphBuilder) -> None:
    roundtrip(GcpAcceleratorType, random_builder)


def test_gcp_address(random_builder: GraphBuilder) -> None:
    address = roundtrip(GcpAddress, random_builder)
    connect_resource(random_builder, address, GcpSubnetwork, selfLink=address.subnetwork)
    assert len(random_builder.edges_of(GcpSubnetwork, GcpAddress)) == 1


def test_gcp_autoscaler(random_builder: GraphBuilder) -> None:
    autoscaler = roundtrip(GcpAutoscaler, random_builder)
    connect_resource(random_builder, autoscaler, GcpInstanceGroupManager, selfLink=autoscaler.autoscaler_target)
    assert len(random_builder.edges_of(GcpAutoscaler, GcpInstanceGroupManager)) == 1


def test_gcp_backend_bucket(random_builder: GraphBuilder) -> None:
    roundtrip(GcpBackendBucket, random_builder)


def test_gcp_backend_service(random_builder: GraphBuilder) -> None:
    service = roundtrip(GcpBackendService, random_builder)
    assert service.health_checks
    assert service.backend_service_backends
    connect_resource(random_builder, service, GcpHealthCheck, selfLink=service.health_checks[0])
    assert len(random_builder.edges_of(GcpBackendService, GcpHealthCheck)) == 1
    connect_resource(random_builder, service, GcpInstanceGroup, selfLink=service.backend_service_backends[0].group)
    assert len(random_builder.edges_of(GcpBackendService, GcpInstanceGroup)) == 1
    connect_resource(random_builder, service, GcpNetwork, selfLink=service.network)
    assert len(random_builder.edges_of(GcpNetwork, GcpBackendService)) == 1


def test_gcp_disk_type(random_builder: GraphBuilder) -> None:
    roundtrip(GcpDiskType, random_builder)


def test_disk_type_ondemand_cost(random_builder: GraphBuilder) -> None:
    known_prices_per_gig = [
        ("pd-standard", "us-east1", 0.08),
    ]
    with open(os.path.dirname(__file__) + "/files/skus.json") as f:
        for r in GcpSku.collect(raw=json.load(f)["skus"], builder=random_builder):
            r.post_process_instance(random_builder, {})

    with open(os.path.dirname(__file__) + "/files/disk_type.json") as f:
        for r in GcpDiskType.collect(raw=json.load(f)["items"]["diskTypes"], builder=random_builder):
            r.post_process_instance(random_builder, {})

    regions = random_builder.resources_of(GcpRegion)
    disk_types = random_builder.resources_of(GcpDiskType)

    for price in known_prices_per_gig:
        region = next((obj for obj in regions if obj.id == price[1]), None)
        disk_type = next((obj for obj in disk_types if obj.name == price[0]), None)
        assert disk_type
        disk_type._region = region
        disk_type.connect_in_graph(random_builder, {"Dummy": "Source"})
        assert disk_type.ondemand_cost is not None
        assert disk_type.ondemand_cost > 0.0
        assert round(disk_type.ondemand_cost, 5) == price[2]


def test_gcp_disk(random_builder: GraphBuilder) -> None:
    disk = roundtrip(GcpDisk, random_builder)
    connect_resource(random_builder, disk, GcpDiskType, selfLink=disk.volume_type)
    assert len(random_builder.edges_of(GcpDiskType, GcpDisk)) == 1


def test_gcp_external_vpn_gateway(random_builder: GraphBuilder) -> None:
    roundtrip(GcpExternalVpnGateway, random_builder)


def test_gcp_firewall_policy(random_builder: GraphBuilder) -> None:
    policy = roundtrip(GcpFirewallPolicy, random_builder)
    assert policy.firewall_policy_rules
    assert policy.firewall_policy_rules[0].target_resources
    connect_resource(random_builder, policy, GcpNetwork, selfLink=policy.firewall_policy_rules[0].target_resources[0])
    assert len(random_builder.edges_of(GcpFirewallPolicy, GcpNetwork)) == 1


def test_gcp_firewall(random_builder: GraphBuilder) -> None:
    firewall = roundtrip(GcpFirewall, random_builder)
    connect_resource(random_builder, firewall, GcpNetwork, selfLink=firewall.network)
    assert len(random_builder.edges_of(GcpFirewall, GcpNetwork)) == 1


def test_gcp_forwarding_rule(random_builder: GraphBuilder) -> None:
    roundtrip(GcpForwardingRule, random_builder)


def test_gcp_network_endpoint_group(random_builder: GraphBuilder) -> None:
    roundtrip(GcpNetworkEndpointGroup, random_builder)


def test_gcp_operation(random_builder: GraphBuilder) -> None:
    roundtrip(GcpOperation, random_builder)


def test_gcp_public_delegated_prefix(random_builder: GraphBuilder) -> None:
    roundtrip(GcpPublicDelegatedPrefix, random_builder)


def test_gcp_health_check(random_builder: GraphBuilder) -> None:
    roundtrip(GcpHealthCheck, random_builder)


def test_gcp_http_health_check(random_builder: GraphBuilder) -> None:
    roundtrip(GcpHttpHealthCheck, random_builder)


def test_gcp_https_health_check(random_builder: GraphBuilder) -> None:
    roundtrip(GcpHttpsHealthCheck, random_builder)


def test_gcp_image(random_builder: GraphBuilder) -> None:
    roundtrip(GcpImage, random_builder)


def test_gcp_instance_group_manager(random_builder: GraphBuilder) -> None:
    roundtrip(GcpInstanceGroupManager, random_builder)


def test_gcp_instance_group(random_builder: GraphBuilder) -> None:
    roundtrip(GcpInstanceGroup, random_builder)


def test_gcp_instance_template(random_builder: GraphBuilder) -> None:
    roundtrip(GcpInstanceTemplate, random_builder)


def test_gcp_instance(random_builder: GraphBuilder) -> None:
    gcp_instance = roundtrip(GcpInstance, random_builder)
    connect_resource(random_builder, gcp_instance, GcpMachineType, selfLink=gcp_instance.machine_type)
    assert len(random_builder.edges_of(GcpMachineType, GcpInstance)) == 1


def test_gcp_instance_custom_machine_type(random_builder: GraphBuilder) -> None:
    CUSTOM_MACHINE_TYPE_PART = "/zones/us-east1-b/machineTypes/e2-custom-medium-1024"
    CUSTOM_MACHINE_TYPE_FULL = (
        f"https://www.googleapis.com/compute/v1/projects/{random_builder.project.id}/{CUSTOM_MACHINE_TYPE_PART}"
    )
    fixture_replies = {
        "Instance": {"machineType": lambda: CUSTOM_MACHINE_TYPE_PART},
        "MachineType": {"selfLink": lambda: CUSTOM_MACHINE_TYPE_FULL},
    }
    assert len(random_builder.resources_of(GcpMachineType)) == 0

    with FixturedClient(random_builder, fixture_replies) as random_builder:
        res: List[GcpInstance] = GcpInstance.collect_resources(random_builder)  # type: ignore
        random_builder.executor.wait_for_submitted_work()
        for node, data in random_builder.graph.nodes(data=True):
            node.connect_in_graph(random_builder, data.get("source") or {})
        first_instance: GcpInstance = res[0]

    assert len(random_builder.resources_of(GcpMachineType)) == 1
    only_machine_type = random_builder.resources_of(GcpMachineType)[0]
    assert first_instance.instance_cores == only_machine_type.instance_cores
    assert first_instance.instance_memory == only_machine_type.instance_memory
    assert only_machine_type._zone
    assert only_machine_type._region


def test_machine_type_ondemand_cost(random_builder: GraphBuilder) -> None:
    # Cross-checking with pricing calculated on https://gcpinstances.doit-intl.com/
    known_prices_linux_ondemand_hourly = [
        ("n2d-standard-8", "us-east1", 0.33797),
        ("f1-micro", "us-east1", 0.00760),
        ("m1-ultramem-160", "us-east1", 25.17240),
        ("a2-megagpu-16g", "us-east1", 8.79698),
        ("c2d-highcpu-112", "europe-west3", 5.40826),
        ("m2-ultramem-416", "us-east1", 74.5344),
        ("m3-megamem-64", "europe-west3", 9.28266),
        ("t2d-standard-16", "europe-west3", 0.87083),
        # TODO complete test cases (c3 missing)
    ]
    with open(os.path.dirname(__file__) + "/files/skus.json") as f:
        GcpSku.collect(raw=json.load(f)["skus"], builder=random_builder)

    with open(os.path.dirname(__file__) + "/files/machine_type.json") as f:
        GcpMachineType.collect(raw=json.load(f)["items"]["machineTypes"], builder=random_builder)

    regions = random_builder.resources_of(GcpRegion)
    machine_types = random_builder.resources_of(GcpMachineType)

    for price in known_prices_linux_ondemand_hourly:
        region = next((obj for obj in regions if obj.id == price[1]), None)
        machine_type = next((obj for obj in machine_types if obj.name == price[0]), None)
        assert machine_type
        machine_type._region = region
        machine_type.post_process_instance(random_builder, {"Dummy": "Source"})
        assert machine_type.ondemand_cost
        assert round(machine_type.ondemand_cost, 5) == price[2]


def test_gcp_interconnect_attachment(random_builder: GraphBuilder) -> None:
    roundtrip(GcpInterconnectAttachment, random_builder)


def test_gcp_interconnect_location(random_builder: GraphBuilder) -> None:
    roundtrip(GcpInterconnectLocation, random_builder)


def test_gcp_interconnect(random_builder: GraphBuilder) -> None:
    roundtrip(GcpInterconnect, random_builder)


def test_gcp_license(random_builder: GraphBuilder) -> None:
    roundtrip(GcpLicense, random_builder)


def test_gcp_machine_image(random_builder: GraphBuilder) -> None:
    roundtrip(GcpMachineImage, random_builder)


def test_gcp_machine_type(random_builder: GraphBuilder) -> None:
    roundtrip(GcpMachineType, random_builder)


def test_gcp_network_edge_security_service(random_builder: GraphBuilder) -> None:
    roundtrip(GcpNetworkEdgeSecurityService, random_builder)


def test_gcp_network(random_builder: GraphBuilder) -> None:
    roundtrip(GcpNetwork, random_builder)


def test_gcp_node_group(random_builder: GraphBuilder) -> None:
    roundtrip(GcpNodeGroup, random_builder)


def test_gcp_node_template(random_builder: GraphBuilder) -> None:
    roundtrip(GcpNodeTemplate, random_builder)


def test_gcp_node_type(random_builder: GraphBuilder) -> None:
    roundtrip(GcpNodeType, random_builder)


def test_gcp_packet_mirroring(random_builder: GraphBuilder) -> None:
    roundtrip(GcpPacketMirroring, random_builder)


def test_gcp_public_advertised_prefix(random_builder: GraphBuilder) -> None:
    roundtrip(GcpPublicAdvertisedPrefix, random_builder)


def test_gcp_commitment(random_builder: GraphBuilder) -> None:
    roundtrip(GcpCommitment, random_builder)


def test_gcp_health_check_service(random_builder: GraphBuilder) -> None:
    roundtrip(GcpHealthCheckService, random_builder)


def test_gcp_notification_endpoint(random_builder: GraphBuilder) -> None:
    roundtrip(GcpNotificationEndpoint, random_builder)


def test_gcp_security_policy(random_builder: GraphBuilder) -> None:
    roundtrip(GcpSecurityPolicy, random_builder)


def test_gcp_ssl_certificate(random_builder: GraphBuilder) -> None:
    roundtrip(GcpSslCertificate, random_builder)


def test_gcp_ssl_policy(random_builder: GraphBuilder) -> None:
    roundtrip(GcpSslPolicy, random_builder)


def test_gcp_target_http_proxy(random_builder: GraphBuilder) -> None:
    roundtrip(GcpTargetHttpProxy, random_builder)


def test_gcp_target_https_proxy(random_builder: GraphBuilder) -> None:
    roundtrip(GcpTargetHttpsProxy, random_builder)


def test_gcp_target_tcp_proxy(random_builder: GraphBuilder) -> None:
    roundtrip(GcpTargetTcpProxy, random_builder)


def test_gcp_url_map(random_builder: GraphBuilder) -> None:
    roundtrip(GcpUrlMap, random_builder)


def test_gcp_resource_policy(random_builder: GraphBuilder) -> None:
    roundtrip(GcpResourcePolicy, random_builder)


def test_gcp_router(random_builder: GraphBuilder) -> None:
    roundtrip(GcpRouter, random_builder)


def test_gcp_route(random_builder: GraphBuilder) -> None:
    roundtrip(GcpRoute, random_builder)


def test_gcp_service_attachment(random_builder: GraphBuilder) -> None:
    roundtrip(GcpServiceAttachment, random_builder)


def test_gcp_snapshot(random_builder: GraphBuilder) -> None:
    roundtrip(GcpSnapshot, random_builder)


def test_gcp_subnetwork(random_builder: GraphBuilder) -> None:
    roundtrip(GcpSubnetwork, random_builder)


def test_gcp_target_grpc_proxy(random_builder: GraphBuilder) -> None:
    roundtrip(GcpTargetGrpcProxy, random_builder)


def test_gcp_target_instance(random_builder: GraphBuilder) -> None:
    roundtrip(GcpTargetInstance, random_builder)


def test_gcp_target_pool(random_builder: GraphBuilder) -> None:
    roundtrip(GcpTargetPool, random_builder)


def test_gcp_target_ssl_proxy(random_builder: GraphBuilder) -> None:
    roundtrip(GcpTargetSslProxy, random_builder)


def test_gcp_target_vpn_gateway(random_builder: GraphBuilder) -> None:
    roundtrip(GcpTargetVpnGateway, random_builder)


def test_gcp_vpn_gateway(random_builder: GraphBuilder) -> None:
    roundtrip(GcpVpnGateway, random_builder)


def test_gcp_vpn_tunnel(random_builder: GraphBuilder) -> None:
    roundtrip(GcpVpnTunnel, random_builder)



import os
import socket
from datetime import datetime
from typing import Iterable, List, Union, Callable, Any, Dict, Optional

from google.oauth2 import service_account
from googleapiclient import discovery
from googleapiclient.discovery_cache.base import Cache as GoogleApiClientCache
from googleapiclient.errors import HttpError as GoogleApiClientHttpError
from retrying import retry
from tenacity import Retrying, stop_after_attempt, retry_if_exception_type

import fixlib.logger
from fixlib.baseresources import BaseResource
from fixlib.config import Config
from fixlib.core.actions import CoreFeedback
from fixlib.graph import Graph
from fixlib.lock import RWLock

log = fixlib.logger.getLogger("fix." + __name__)
fixlib.logger.getLogger("googleapiclient").setLevel(fixlib.logger.ERROR)


SCOPES = ["https://www.googleapis.com/auth/cloud-platform"]


def retry_on_error(e):
    if isinstance(e, socket.timeout):
        log.debug("Got socket timeout - retrying")
        return True
    return False


class MemoryCache(GoogleApiClientCache):
    _cache = {}

    def get(self, url):
        return MemoryCache._cache.get(url)

    def set(self, url, content):
        MemoryCache._cache[url] = content


class Credentials:
    _credentials = {}
    _initialized = False
    _lock = RWLock()

    @staticmethod
    def load(feedback: Optional[CoreFeedback] = None):
        with Credentials._lock.write_access:
            if not Credentials._initialized:
                for sa_data in Config.gcp.service_account:
                    try:
                        log.debug("Loading credentials from %s", sa_data)
                        credentials = load_credentials(sa_data)
                        for project in list_credential_projects(credentials):
                            Credentials._credentials[project["id"]] = credentials
                    except Exception as e:
                        log.error("Unable to load credentials from %s", sa_data, exc_info=e)
                        if feedback is not None:
                            feedback.error(f"Unable to load credentials from {sa_data}: {e}")
                Credentials._initialized = True

    @staticmethod
    def get(project_id: str):
        Credentials.load()
        with Credentials._lock.read_access:
            return Credentials._credentials.get(project_id)

    @staticmethod
    def all(feedback: Optional[CoreFeedback] = None) -> Dict:
        Credentials.load(feedback)
        with Credentials._lock.read_access:
            return dict(Credentials._credentials)

    @staticmethod
    def reload():
        with Credentials._lock.write_access:
            Credentials._initialized = False
        Credentials.load()


def load_credentials(path: str):
    if len(path) == 0:
        return None
    file = os.path.expanduser(path)
    if os.path.isfile(file):
        return service_account.Credentials.from_service_account_file(file, scopes=SCOPES)
    else:
        raise ValueError(f"No credentials file found at {file}")


@retry(
    stop_max_attempt_number=10,
    wait_exponential_multiplier=3000,
    wait_exponential_max=300000,
    retry_on_exception=retry_on_error,
)
def gcp_client(service: str, version: str, credentials: str):
    client = discovery.build(service, version, credentials=credentials, cache=MemoryCache())
    return client


def list_credential_projects(credentials) -> List:
    ret = []
    try:
        client = gcp_client("cloudresourcemanager", "v1", credentials=credentials)
        projects = client.projects()
        for project in paginate(projects, "list", "projects"):
            ctime = project.get("createTime")
            if ctime is not None:
                ctime = iso2datetime(ctime)
            project_name = project.get("name")
            project_id = project.get("projectId", "")
            if project_id.startswith("sys-"):
                continue
            p = {
                "id": project_id,
                "name": project_name,
                "ctime": ctime,
            }
            ret.append(p)
    except GoogleApiClientHttpError:
        log.error("Unable to load projects from cloudresourcemanager - falling back to local credentials information")
        p = {
            "id": credentials.project_id,
            "name": credentials.project_id,
        }
        ret.append(p)
    return ret


def iso2datetime(ts: str) -> datetime:
    if ts is None:
        return
    if ts.endswith("Z"):
        ts = ts[:-1] + "+00:00"
    if ts is not None:
        return datetime.fromisoformat(ts)


def paginate(
    gcp_resource: Callable,
    method_name: str,
    items_name: str,
    subitems_name: str = None,
    exclude_region_resources: bool = False,
    **kwargs,
) -> Iterable:
    """Paginate GCP API list and aggregatedList results.

    Args:
        gcp_resource: GCP resource on which we do our paging
        method_name: list method to call. Usually `list` or `aggregatedList`
        items_name: Name of the key in our result that contains the list of items.
            Usually `items`
        subitems_name: When using aggregatedList this contains the actual items.
            Usually the same as the gcp_resource name. E.g. `disks` when requesting
            disks, `instances` when fetching instances, etc.
        exclude_region_resources: Regional resources have their own API and can be
            excluded from aggregatedList calls if so desired
    """
    next_method_name = method_name + "_next"
    method = getattr(gcp_resource, method_name)
    request = method(**kwargs)
    while request is not None:
        for attempt in Retrying(
            reraise=True,
            stop=stop_after_attempt(10),
            retry=retry_if_exception_type(socket.timeout),
        ):
            with attempt:
                result = request.execute()
        if items_name in result:
            items = result[items_name]
            if isinstance(items, dict):
                for location, item in items.items():
                    if (
                        method_name == "aggregatedList"
                        and exclude_region_resources
                        and str(location).startswith("regions/")
                    ):
                        continue
                    if subitems_name in item:
                        yield from item[subitems_name]
            else:
                yield from items
        if hasattr(gcp_resource, next_method_name):
            method = getattr(gcp_resource, next_method_name)
            request = method(request, result)
        else:
            request = None


def get_result_data(result: Dict, value: Union[str, Callable]) -> Any:
    """Returns data from a GCP API call result dict.

    Args:
        result: Dict containing the result or a GCP API execute() call.
        value: Either directly the name of a key found in result or
            a callable like a lambda that finds the relevant data withing
            result.
    """
    data = None
    if callable(value):
        try:
            data = value(result)
        except Exception:
            log.exception(f"Exception while trying to fetch data calling {value}")
    elif value in result:
        data = result[value]
    return data


def common_resource_kwargs(resource: BaseResource) -> Dict:
    common_kwargs = {}
    if resource.account().id != "undefined" and "project" in resource.resource_args:
        common_kwargs["project"] = resource.account().id
    if resource.zone().name != "undefined" and "zone" in resource.resource_args:
        common_kwargs["zone"] = resource.zone().name
    elif resource.region().name != "undefined" and "region" in resource.resource_args:
        common_kwargs["region"] = resource.region().name
    return common_kwargs


def delete_resource(resource: BaseResource) -> bool:
    delete_kwargs = {str(resource._delete_identifier): resource.name}
    common_kwargs = common_resource_kwargs(resource)
    delete_kwargs.update(common_kwargs)

    gr = gcp_resource(resource)
    request = gr.delete(**delete_kwargs)
    request.execute()
    return True


def gcp_service(resource: BaseResource, graph: Graph = None):
    service_kwargs = {}
    if resource.account().id != "undefined":
        service_kwargs["credentials"] = Credentials.get(resource.account(graph).id)
    return gcp_client(resource.client, resource.api_version, **service_kwargs)


def gcp_resource(resource: BaseResource, graph: Graph = None):
    service = gcp_service(resource, graph)
    gr = getattr(service, resource._client_method)
    return gr()



from datetime import datetime
from typing import ClassVar, Dict, Optional, List, Type, cast

from attr import define, field

from fix_plugin_gcp.gcp_client import GcpApiSpec
from fix_plugin_gcp.resources.base import GcpResource, GcpDeprecationStatus, GraphBuilder
from fixlib.baseresources import ModelReference
from fixlib.json_bender import Bender, S, Bend, ForallBend
from fixlib.types import Json

# This service is called Cloud Billing in the documentation
# https://cloud.google.com/billing/docs
# API https://googleapis.github.io/google-api-python-client/docs/dyn/cloudbilling_v1.html


@define(eq=False, slots=False)
class GcpBillingAccount(GcpResource):
    kind: ClassVar[str] = "gcp_billing_account"
    kind_display: ClassVar[str] = "GCP Billing Account"
    kind_description: ClassVar[str] = (
        "GCP Billing Account is a financial account used to manage the payment and"
        " billing information for Google Cloud Platform services."
    )
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["gcp_project_billing_info"]},
    }
    api_spec: ClassVar[GcpApiSpec] = GcpApiSpec(
        service="cloudbilling",
        version="v1",
        accessors=["billingAccounts"],
        action="list",
        request_parameter={},
        request_parameter_in=set(),
        response_path="billingAccounts",
        response_regional_sub_path=None,
        required_iam_permissions=[],  # does not require any permissions
        mutate_iam_permissions=[],  # can not be deleted
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("name").or_else(S("id")).or_else(S("selfLink")),
        "tags": S("labels", default={}),
        "name": S("name"),
        "ctime": S("creationTimestamp"),
        "description": S("description"),
        "link": S("selfLink"),
        "label_fingerprint": S("labelFingerprint"),
        "deprecation_status": S("deprecated", default={}) >> Bend(GcpDeprecationStatus.mapping),
        "display_name": S("displayName"),
        "master_billing_account": S("masterBillingAccount"),
        "open": S("open"),
    }

    display_name: Optional[str] = field(default=None)
    master_billing_account: Optional[str] = field(default=None)
    open: Optional[bool] = field(default=None)

    def post_process(self, graph_builder: GraphBuilder, source: Json) -> None:
        for info in GcpProjectBillingInfo.collect_resources(graph_builder, name=self.name):
            graph_builder.add_edge(self, node=info)

    @classmethod
    def called_collect_apis(cls) -> List[GcpApiSpec]:
        return [cls.api_spec, GcpProjectBillingInfo.api_spec]


@define(eq=False, slots=False)
class GcpProjectBillingInfo(GcpResource):
    kind: ClassVar[str] = "gcp_project_billing_info"
    kind_display: ClassVar[str] = "GCP Project Billing Info"
    kind_description: ClassVar[str] = (
        "GCP Project Billing Info provides information and management capabilities"
        " for the billing aspects of a Google Cloud Platform project."
    )
    api_spec: ClassVar[GcpApiSpec] = GcpApiSpec(
        service="cloudbilling",
        version="v1",
        accessors=["billingAccounts", "projects"],
        action="list",
        request_parameter={"name": "{name}"},
        request_parameter_in={"name"},
        response_path="projectBillingInfo",
        response_regional_sub_path=None,
        # valid permission name according to documentation, but gcloud emits an error
        # required_iam_permissions=["billing.resourceAssociations.list"],
        required_iam_permissions=[],
        mutate_iam_permissions=["billing.resourceAssociations.delete"],
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("name").or_else(S("id")).or_else(S("selfLink")),
        "tags": S("labels", default={}),
        "name": S("name"),
        "ctime": S("creationTimestamp"),
        "description": S("description"),
        "link": S("selfLink"),
        "label_fingerprint": S("labelFingerprint"),
        "deprecation_status": S("deprecated", default={}) >> Bend(GcpDeprecationStatus.mapping),
        "billing_account_name": S("billingAccountName"),
        "billing_enabled": S("billingEnabled"),
        "project_billing_info_project_id": S("projectId"),
    }
    billing_account_name: Optional[str] = field(default=None)
    billing_enabled: Optional[bool] = field(default=None)
    project_billing_info_project_id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpService(GcpResource):
    kind: ClassVar[str] = "gcp_service"
    kind_display: ClassVar[str] = "GCP Service"
    kind_description: ClassVar[str] = (
        "GCP Service refers to any of the various services and products offered by"
        " Google Cloud Platform, which provide scalable cloud computing solutions for"
        " businesses and developers."
    )
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {"default": ["gcp_sku"]},
    }
    api_spec: ClassVar[GcpApiSpec] = GcpApiSpec(
        service="cloudbilling",
        version="v1",
        accessors=["services"],
        action="list",
        request_parameter={},
        request_parameter_in=set(),
        response_path="services",
        response_regional_sub_path=None,
        required_iam_permissions=[],  # does not require any permissions
        mutate_iam_permissions=[],  # can not be deleted
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("serviceId"),
        "tags": S("labels", default={}),
        "name": S("name"),
        "display_name": S("displayName"),
        "ctime": S("creationTimestamp"),
        "description": S("description"),
        "link": S("selfLink"),
        "label_fingerprint": S("labelFingerprint"),
        "deprecation_status": S("deprecated", default={}) >> Bend(GcpDeprecationStatus.mapping),
        "business_entity_name": S("businessEntityName"),
    }

    business_entity_name: Optional[str] = field(default=None)
    display_name: Optional[str] = field(default=None)

    @classmethod
    def collect(cls: Type[GcpResource], raw: List[Json], builder: GraphBuilder) -> List[GcpResource]:
        # Additional behavior: iterate over list of collected GcpService and for each:
        # - collect related GcpSku
        result: List[GcpResource] = super().collect(raw, builder)  # type: ignore
        SERVICES_COLLECT_LIST = [
            "Compute Engine",
        ]
        service_names = [
            service.name for service in cast(List[GcpService], result) if service.display_name in SERVICES_COLLECT_LIST
        ]
        for service_name in service_names:
            builder.submit_work(GcpSku.collect_resources, builder, parent=service_name)

        return result

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        def filter(node: GcpResource) -> bool:
            return isinstance(node, GcpSku) and node.name is not None and node.name.startswith(self.id)

        builder.add_edges(self, filter=filter)

    @classmethod
    def called_collect_apis(cls) -> List[GcpApiSpec]:
        return [cls.api_spec, GcpSku.api_spec]


@define(eq=False, slots=False)
class GcpCategory:
    kind: ClassVar[str] = "gcp_category"
    kind_display: ClassVar[str] = "GCP Category"
    kind_description: ClassVar[str] = (
        "GCP Category is a classification system used by Google Cloud Platform to"
        " organize various cloud resources and services into different categories."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "resource_family": S("resourceFamily"),
        "resource_group": S("resourceGroup"),
        "service_display_name": S("serviceDisplayName"),
        "usage_type": S("usageType"),
    }
    resource_family: Optional[str] = field(default=None)
    resource_group: Optional[str] = field(default=None)
    service_display_name: Optional[str] = field(default=None)
    usage_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpGeoTaxonomy:
    kind: ClassVar[str] = "gcp_geo_taxonomy"
    kind_display: ClassVar[str] = "GCP Geo Taxonomy"
    kind_description: ClassVar[str] = (
        "GCP Geo Taxonomy within a SKU refers to the classification of Google Cloud resources"
        " based on geographic regions and types, which impacts pricing and availability."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"regions": S("regions", default=[]), "type": S("type")}
    regions: List[str] = field(factory=list)
    type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpAggregationInfo:
    kind: ClassVar[str] = "gcp_aggregation_info"
    kind_display: ClassVar[str] = "GCP Aggregation Info"
    kind_description: ClassVar[str] = (
        "GCP Aggregation Info refers to how usage and cost data are compiled and summarized over time."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "aggregation_count": S("aggregationCount"),
        "aggregation_interval": S("aggregationInterval"),
        "aggregation_level": S("aggregationLevel"),
    }
    aggregation_count: Optional[int] = field(default=None)
    aggregation_interval: Optional[str] = field(default=None)
    aggregation_level: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpMoney:
    kind: ClassVar[str] = "gcp_money"
    kind_display: ClassVar[str] = "GCP Money"
    kind_description: ClassVar[str] = (
        "In GCP's Money structure, amounts are represented using a currency_code for the type of currency,"
        " and nanos to denote a fraction of that currency down to one-billionth, ensuring precise"
        " financial calculations."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "currency_code": S("currencyCode"),
        "nanos": S("nanos"),
        "units": S("units"),
    }
    currency_code: Optional[str] = field(default=None)
    nanos: Optional[int] = field(default=None)
    units: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpTierRate:
    kind: ClassVar[str] = "gcp_tier_rate"
    kind_display: ClassVar[str] = "GCP Tier Rate"
    kind_description: ClassVar[str] = (
        "GCP Tier Rate refers to the pricing tiers for different levels of usage of"
        " Google Cloud Platform services. Higher tiers typically offer discounted"
        " rates for increased usage volumes."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "start_usage_amount": S("startUsageAmount"),
        "unit_price": S("unitPrice", default={}) >> Bend(GcpMoney.mapping),
    }
    start_usage_amount: Optional[float] = field(default=None)
    unit_price: Optional[GcpMoney] = field(default=None)


@define(eq=False, slots=False)
class GcpPricingExpression:
    kind: ClassVar[str] = "gcp_pricing_expression"
    kind_display: ClassVar[str] = "GCP Pricing Expression"
    kind_description: ClassVar[str] = (
        "GCP Pricing Expression delineates the structure of pricing for a particular service, including the base"
        " units of measurement, conversion factors, detailed descriptions, and tiered pricing rates to calculate"
        " the cost based on usage."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "base_unit": S("baseUnit"),
        "base_unit_conversion_factor": S("baseUnitConversionFactor"),
        "base_unit_description": S("baseUnitDescription"),
        "display_quantity": S("displayQuantity"),
        "tiered_rates": S("tieredRates", default=[]) >> ForallBend(GcpTierRate.mapping),
        "usage_unit": S("usageUnit"),
        "usage_unit_description": S("usageUnitDescription"),
    }
    base_unit: Optional[str] = field(default=None)
    base_unit_conversion_factor: Optional[float] = field(default=None)
    base_unit_description: Optional[str] = field(default=None)
    display_quantity: Optional[float] = field(default=None)
    tiered_rates: List[GcpTierRate] = field(factory=list)
    usage_unit: Optional[str] = field(default=None)
    usage_unit_description: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpPricingInfo:
    kind: ClassVar[str] = "gcp_pricing_info"
    kind_display: ClassVar[str] = "GCP Pricing Info"
    kind_description: ClassVar[str] = (
        "GCP Pricing Info provides information on the pricing models and costs"
        " associated with using Google Cloud Platform services."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "aggregation_info": S("aggregationInfo", default={}) >> Bend(GcpAggregationInfo.mapping),
        "currency_conversion_rate": S("currencyConversionRate"),
        "effective_time": S("effectiveTime"),
        "pricing_expression": S("pricingExpression", default={}) >> Bend(GcpPricingExpression.mapping),
        "summary": S("summary"),
    }
    aggregation_info: Optional[GcpAggregationInfo] = field(default=None)
    currency_conversion_rate: Optional[float] = field(default=None)
    effective_time: Optional[datetime] = field(default=None)
    pricing_expression: Optional[GcpPricingExpression] = field(default=None)
    summary: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpSku(GcpResource):
    kind: ClassVar[str] = "gcp_sku"
    kind_display: ClassVar[str] = "GCP SKU"
    kind_description: ClassVar[str] = (
        "GCP SKU represents a Stock Keeping Unit in Google Cloud Platform, providing"
        " unique identifiers for different resources and services."
    )
    api_spec: ClassVar[GcpApiSpec] = GcpApiSpec(
        service="cloudbilling",
        version="v1",
        accessors=["services", "skus"],
        action="list",
        request_parameter={"parent": "{parent}"},
        request_parameter_in={"parent"},
        response_path="skus",
        response_regional_sub_path=None,
        required_iam_permissions=[],  # does not require any permissions
        mutate_iam_permissions=[],  # can not be deleted
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("skuId"),
        "tags": S("labels", default={}),
        "name": S("name"),
        "ctime": S("creationTimestamp"),
        "description": S("description"),
        "link": S("selfLink"),
        "label_fingerprint": S("labelFingerprint"),
        "deprecation_status": S("deprecated", default={}) >> Bend(GcpDeprecationStatus.mapping),
        "category": S("category", default={}) >> Bend(GcpCategory.mapping),
        "geo_taxonomy": S("geoTaxonomy", default={}) >> Bend(GcpGeoTaxonomy.mapping),
        "sku_pricing_info": S("pricingInfo", default=[]) >> ForallBend(GcpPricingInfo.mapping),
        "service_provider_name": S("serviceProviderName"),
        "service_regions": S("serviceRegions", default=[]),
        "sku_id": S("skuId"),
    }
    category: Optional[GcpCategory] = field(default=None)
    geo_taxonomy: Optional[GcpGeoTaxonomy] = field(default=None)
    sku_pricing_info: List[GcpPricingInfo] = field(factory=list)
    service_provider_name: Optional[str] = field(default=None)
    service_regions: List[str] = field(factory=list)
    usage_unit_nanos: Optional[int] = field(default=None)

    def post_process(self, graph_builder: GraphBuilder, source: Json) -> None:
        if len(self.sku_pricing_info) > 0:
            if not (pricing_expression := self.sku_pricing_info[0].pricing_expression):
                return

            tiered_rates = pricing_expression.tiered_rates
            cost = -1
            if len(tiered_rates) == 1:
                if tiered_rates[0].unit_price and tiered_rates[0].unit_price.nanos:
                    cost = tiered_rates[0].unit_price.nanos

            else:
                for tiered_rate in tiered_rates:
                    if sua := tiered_rate.start_usage_amount:
                        if sua > 0:
                            if tiered_rate.unit_price and tiered_rate.unit_price.nanos:
                                cost = tiered_rate.unit_price.nanos
                                break
            if cost > -1:
                self.usage_unit_nanos = cost


resources = [GcpBillingAccount, GcpService]



from datetime import datetime
from typing import ClassVar, Dict, Optional, List

from attr import define, field

from fix_plugin_gcp.gcp_client import GcpApiSpec
from fix_plugin_gcp.resources.base import GcpResource, GcpDeprecationStatus, get_client
from fixlib.graph import Graph
from fixlib.json_bender import Bender, S, Bend, ForallBend


@define(eq=False, slots=False)
class GcpProjectteam:
    kind: ClassVar[str] = "gcp_projectteam"
    kind_display: ClassVar[str] = "GCP Project Team"
    kind_description: ClassVar[str] = (
        "GCP Project Teams are groups of users who work together on a Google Cloud"
        " Platform project, allowing them to collaborate and manage resources within"
        " the project."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"project_number": S("projectNumber"), "team": S("team")}
    project_number: Optional[str] = field(default=None)
    team: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpBucketAccessControl:
    kind: ClassVar[str] = "gcp_bucket_access_control"
    kind_display: ClassVar[str] = "GCP Bucket Access Control"
    kind_description: ClassVar[str] = (
        "Bucket Access Control is a feature in the Google Cloud Platform that allows"
        " you to manage and control access to your storage buckets. It provides fine-"
        " grained access control, allowing you to specify who can read, write, or"
        " delete objects within a bucket."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("name").or_else(S("id")).or_else(S("selfLink")),
        "tags": S("labels", default={}),
        "name": S("name"),
        "ctime": S("creationTimestamp"),
        "description": S("description"),
        "link": S("selfLink"),
        "label_fingerprint": S("labelFingerprint"),
        "deprecation_status": S("deprecated", default={}) >> Bend(GcpDeprecationStatus.mapping),
        "bucket": S("bucket"),
        "domain": S("domain"),
        "email": S("email"),
        "entity": S("entity"),
        "entity_id": S("entityId"),
        "etag": S("etag"),
        "project_team": S("projectTeam", default={}) >> Bend(GcpProjectteam.mapping),
        "role": S("role"),
    }
    bucket: Optional[str] = field(default=None)
    domain: Optional[str] = field(default=None)
    email: Optional[str] = field(default=None)
    entity: Optional[str] = field(default=None)
    entity_id: Optional[str] = field(default=None)
    etag: Optional[str] = field(default=None)
    project_team: Optional[GcpProjectteam] = field(default=None)
    role: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpAutoclass:
    kind: ClassVar[str] = "gcp_autoclass"
    kind_display: ClassVar[str] = "GCP Autoclass"
    kind_description: ClassVar[str] = (
        "GCP Autoclass for a bucket indicates if automatic data classification is enabled"
        " and the timestamp when this setting was last changed."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("enabled"), "toggle_time": S("toggleTime")}
    enabled: Optional[bool] = field(default=None)
    toggle_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class GcpCors:
    kind: ClassVar[str] = "gcp_cors"
    kind_display: ClassVar[str] = "GCP CORS"
    kind_description: ClassVar[str] = (
        "GCP CORS settings for a bucket in Google Cloud Storage define which origins can access resources"
        " in the bucket, what HTTP methods are allowed, and which response headers can be exposed to the client."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "max_age_seconds": S("maxAgeSeconds"),
        "method": S("method", default=[]),
        "origin": S("origin", default=[]),
        "response_header": S("responseHeader", default=[]),
    }
    max_age_seconds: Optional[int] = field(default=None)
    method: Optional[List[str]] = field(default=None)
    origin: Optional[List[str]] = field(default=None)
    response_header: Optional[List[str]] = field(default=None)


@define(eq=False, slots=False)
class GcpObjectAccessControl:
    kind: ClassVar[str] = "gcp_object_access_control"
    kind_display: ClassVar[str] = "GCP Object Access Control"
    kind_description: ClassVar[str] = (
        "GCP Object Access Control is a feature in Google Cloud Platform that allows"
        " users to control access to their storage objects (such as files, images,"
        " videos) by specifying permissions and policies."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "bucket": S("bucket"),
        "domain": S("domain"),
        "email": S("email"),
        "entity": S("entity"),
        "entity_id": S("entityId"),
        "etag": S("etag"),
        "generation": S("generation"),
        "id": S("id"),
        "object": S("object"),
        "project_team": S("projectTeam", default={}) >> Bend(GcpProjectteam.mapping),
        "role": S("role"),
        "self_link": S("selfLink"),
    }
    bucket: Optional[str] = field(default=None)
    domain: Optional[str] = field(default=None)
    email: Optional[str] = field(default=None)
    entity: Optional[str] = field(default=None)
    entity_id: Optional[str] = field(default=None)
    etag: Optional[str] = field(default=None)
    generation: Optional[str] = field(default=None)
    id: Optional[str] = field(default=None)
    object: Optional[str] = field(default=None)
    project_team: Optional[GcpProjectteam] = field(default=None)
    role: Optional[str] = field(default=None)
    self_link: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpBucketpolicyonly:
    kind: ClassVar[str] = "gcp_bucketpolicyonly"
    kind_display: ClassVar[str] = "GCP Bucket Policy Only"
    kind_description: ClassVar[str] = (
        "GCP Bucket Policy Only is a feature in Google Cloud Platform that enforces"
        " the use of IAM policies for bucket access control and disables any ACL-based"
        " access control."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("enabled"), "locked_time": S("lockedTime")}
    enabled: Optional[bool] = field(default=None)
    locked_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class GcpUniformbucketlevelaccess:
    kind: ClassVar[str] = "gcp_uniformbucketlevelaccess"
    kind_display: ClassVar[str] = "GCP Uniform Bucket Level Access"
    kind_description: ClassVar[str] = (
        "GCP Uniform Bucket Level Access is a Google Cloud Storage feature that, when enabled, applies consistent"
        " IAM policies across all objects in a bucket, eliminating the need for individual object permissions."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("enabled"), "locked_time": S("lockedTime")}
    enabled: Optional[bool] = field(default=None)
    locked_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class GcpIamconfiguration:
    kind: ClassVar[str] = "gcp_iamconfiguration"
    kind_display: ClassVar[str] = "GCP IAM Configuration"
    kind_description: ClassVar[str] = (
        "IAM (Identity and Access Management) Configuration in Google Cloud Platform,"
        " which allows users to control access to their cloud resources and manage"
        " permissions for users and service accounts."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "bucket_policy_only": S("bucketPolicyOnly", default={}) >> Bend(GcpBucketpolicyonly.mapping),
        "public_access_prevention": S("publicAccessPrevention"),
        "uniform_bucket_level_access": S("uniformBucketLevelAccess", default={})
        >> Bend(GcpUniformbucketlevelaccess.mapping),
    }
    bucket_policy_only: Optional[GcpBucketpolicyonly] = field(default=None)
    public_access_prevention: Optional[str] = field(default=None)
    uniform_bucket_level_access: Optional[GcpUniformbucketlevelaccess] = field(default=None)


@define(eq=False, slots=False)
class GcpAction:
    kind: ClassVar[str] = "gcp_action"
    kind_display: ClassVar[str] = "GCP Action"
    kind_description: ClassVar[str] = (
        "GCP Action refers to a specific action or operation performed on resources"
        " in Google Cloud Platform (GCP), such as creating, deleting, or modifying"
        " cloud resources."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"storage_class": S("storageClass"), "type": S("type")}
    storage_class: Optional[str] = field(default=None)
    type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpCondition:
    kind: ClassVar[str] = "gcp_condition"
    kind_display: ClassVar[str] = "GCP Condition"
    kind_description: ClassVar[str] = (
        "Conditions in Google Cloud Platform (GCP) are used to define rules and"
        " policies for resource usage and access control."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "age": S("age"),
        "created_before": S("createdBefore"),
        "custom_time_before": S("customTimeBefore"),
        "days_since_custom_time": S("daysSinceCustomTime"),
        "days_since_noncurrent_time": S("daysSinceNoncurrentTime"),
        "is_live": S("isLive"),
        "matches_pattern": S("matchesPattern"),
        "matches_prefix": S("matchesPrefix", default=[]),
        "matches_storage_class": S("matchesStorageClass", default=[]),
        "matches_suffix": S("matchesSuffix", default=[]),
        "noncurrent_time_before": S("noncurrentTimeBefore"),
        "num_newer_versions": S("numNewerVersions"),
    }
    age: Optional[int] = field(default=None)
    created_before: Optional[str] = field(default=None)
    custom_time_before: Optional[str] = field(default=None)
    days_since_custom_time: Optional[datetime] = field(default=None)  # should be int
    days_since_noncurrent_time: Optional[datetime] = field(default=None)  # should be int
    is_live: Optional[bool] = field(default=None)
    matches_pattern: Optional[str] = field(default=None)
    matches_prefix: Optional[List[str]] = field(default=None)
    matches_storage_class: Optional[List[str]] = field(default=None)
    matches_suffix: Optional[List[str]] = field(default=None)
    noncurrent_time_before: Optional[str] = field(default=None)
    num_newer_versions: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class GcpRule:
    kind: ClassVar[str] = "gcp_rule"
    kind_display: ClassVar[str] = "GCP Rule"
    kind_description: ClassVar[str] = (
        "A GCP Rule for a Bucket refers to an object lifecycle management rule, which automates the deletion"
        " or transition of objects to less expensive storage classes based on specified conditions and actions."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "action": S("action", default={}) >> Bend(GcpAction.mapping),
        "condition": S("condition", default={}) >> Bend(GcpCondition.mapping),
    }
    action: Optional[GcpAction] = field(default=None)
    condition: Optional[GcpCondition] = field(default=None)


@define(eq=False, slots=False)
class GcpLogging:
    kind: ClassVar[str] = "gcp_logging"
    kind_display: ClassVar[str] = "GCP Logging"
    kind_description: ClassVar[str] = (
        "GCP Logging is a service provided by Google Cloud Platform that allows users"
        " to collect, store, and analyze logs from various resources in their cloud"
        " environment."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"log_bucket": S("logBucket"), "log_object_prefix": S("logObjectPrefix")}
    log_bucket: Optional[str] = field(default=None)
    log_object_prefix: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpOwner:
    kind: ClassVar[str] = "gcp_owner"
    kind_display: ClassVar[str] = "GCP Owner"
    kind_description: ClassVar[str] = (
        "GCP Owner refers to the owner of a Google Cloud Platform (GCP) resource or"
        " project, who has full control and access to manage and configure the"
        " resource or project."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"entity": S("entity"), "entity_id": S("entityId")}
    entity: Optional[str] = field(default=None)
    entity_id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpRetentionpolicy:
    kind: ClassVar[str] = "gcp_retentionpolicy"
    kind_display: ClassVar[str] = "GCP Retention Policy"
    kind_description: ClassVar[str] = (
        "GCP Retention Policy is a feature in Google Cloud Platform that allows users"
        " to set and manage rules for data retention, specifying how long data should"
        " be kept before it is automatically deleted."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "effective_time": S("effectiveTime"),
        "is_locked": S("isLocked"),
        "retention_period": S("retentionPeriod"),
    }
    effective_time: Optional[datetime] = field(default=None)
    is_locked: Optional[bool] = field(default=None)
    retention_period: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpWebsite:
    kind: ClassVar[str] = "gcp_website"
    kind_display: ClassVar[str] = "GCP Website"
    kind_description: ClassVar[str] = (
        "GCP Website refers to the official website of Google Cloud Platform where"
        " users can access information, documentation, and resources related to"
        " Google's cloud services and products."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "main_page_suffix": S("mainPageSuffix"),
        "not_found_page": S("notFoundPage"),
    }
    main_page_suffix: Optional[str] = field(default=None)
    not_found_page: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpObject(GcpResource):
    # GcpObjects are necessary to empty buckets before deletion
    # they are not intended to be collected and stored in the graph
    kind: ClassVar[str] = "gcp_object"
    kind_display: ClassVar[str] = "GCP Object"
    kind_description: ClassVar[str] = (
        "GCP Object, specifically referring to the Google Cloud Storage, is a basic unit of data that is stored"
        " in Google Cloud Storage, often matching to an individual file."
    )
    api_spec: ClassVar[GcpApiSpec] = GcpApiSpec(
        service="storage",
        version="v1",
        accessors=["objects"],
        action="list",
        request_parameter={"bucket": "{bucket}"},
        request_parameter_in={"bucket"},
        response_path="items",
        response_regional_sub_path=None,
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("name").or_else(S("id")).or_else(S("selfLink")),
        "name": S("name"),
    }


@define(eq=False, slots=False)
class GcpBucket(GcpResource):
    kind: ClassVar[str] = "gcp_bucket"
    kind_display: ClassVar[str] = "GCP Bucket"
    kind_description: ClassVar[str] = (
        "A GCP Bucket is a cloud storage container provided by Google Cloud Platform,"
        " allowing users to store and access data in a scalable and durable manner."
    )
    api_spec: ClassVar[GcpApiSpec] = GcpApiSpec(
        service="storage",
        version="v1",
        accessors=["buckets"],
        action="list",
        request_parameter={"project": "{project}"},
        request_parameter_in={"project"},
        # single_request_parameter={"project": "{project}"},
        # single_request_parameter_in={"project"},
        response_path="items",
        response_regional_sub_path=None,
        mutate_iam_permissions=["storage.buckets.update", "storage.buckets.delete"],
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("name").or_else(S("id")).or_else(S("selfLink")),
        "tags": S("labels", default={}),
        "name": S("name"),
        "ctime": S("creationTimestamp"),
        "mtime": S("updated"),
        "description": S("description"),
        "link": S("selfLink"),
        "label_fingerprint": S("labelFingerprint"),
        "deprecation_status": S("deprecated", default={}) >> Bend(GcpDeprecationStatus.mapping),
        "acl": S("acl", default=[]) >> ForallBend(GcpBucketAccessControl.mapping),
        "autoclass": S("autoclass", default={}) >> Bend(GcpAutoclass.mapping),
        "requester_pays": S("billing", "requesterPays"),
        "cors": S("cors", default=[]) >> ForallBend(GcpCors.mapping),
        "custom_placement_config": S("customPlacementConfig", "data_locations", default=[]),
        "default_event_based_hold": S("defaultEventBasedHold"),
        "default_object_acl": S("defaultObjectAcl", default=[]) >> ForallBend(GcpObjectAccessControl.mapping),
        "encryption_default_kms_key_name": S("encryption", "defaultKmsKeyName"),
        "etag": S("etag"),
        "iam_configuration": S("iamConfiguration", default={}) >> Bend(GcpIamconfiguration.mapping),
        "lifecycle_rule": S("lifecycle", "rule", default=[]) >> ForallBend(GcpRule.mapping),
        "location": S("location"),
        "location_type": S("locationType"),
        "logging": S("logging", default={}) >> Bend(GcpLogging.mapping),
        "metageneration": S("metageneration"),
        "bucket_owner": S("owner", default={}) >> Bend(GcpOwner.mapping),
        "project_number": S("projectNumber"),
        "retention_policy": S("retentionPolicy", default={}) >> Bend(GcpRetentionpolicy.mapping),
        "rpo": S("rpo"),
        "satisfies_pzs": S("satisfiesPZS"),
        "storage_class": S("storageClass"),
        "time_created": S("timeCreated"),
        "updated": S("updated"),
        "versioning_enabled": S("versioning", "enabled"),
        "bucket_website": S("website", default={}) >> Bend(GcpWebsite.mapping),
    }
    acl: Optional[List[GcpBucketAccessControl]] = field(default=None)
    autoclass: Optional[GcpAutoclass] = field(default=None)
    cors: Optional[List[GcpCors]] = field(default=None)
    custom_placement_config_data_locations: Optional[List[str]] = field(default=None)
    default_event_based_hold: Optional[bool] = field(default=None)
    default_object_acl: Optional[List[GcpObjectAccessControl]] = field(default=None)
    encryption_default_kms_key_name: Optional[str] = field(default=None)
    etag: Optional[str] = field(default=None)
    iam_configuration: Optional[GcpIamconfiguration] = field(default=None)
    location: Optional[str] = field(default=None)
    location_type: Optional[str] = field(default=None)
    logging: Optional[GcpLogging] = field(default=None)
    metageneration: Optional[str] = field(default=None)
    bucket_owner: Optional[GcpOwner] = field(default=None)
    project_number: Optional[str] = field(default=None)
    retention_policy: Optional[GcpRetentionpolicy] = field(default=None)
    rpo: Optional[str] = field(default=None)
    satisfies_pzs: Optional[bool] = field(default=None)
    storage_class: Optional[str] = field(default=None)
    time_created: Optional[datetime] = field(default=None)
    updated: Optional[datetime] = field(default=None)
    bucket_website: Optional[GcpWebsite] = field(default=None)
    requester_pays: Optional[bool] = field(default=None)
    versioning_enabled: Optional[bool] = field(default=None)
    lifecycle_rule: List[GcpRule] = field(factory=list)

    def pre_delete(self, graph: Graph) -> bool:
        client = get_client(self)
        objects = client.list(GcpObject.api_spec, bucket=self.name)
        for obj in objects:
            object_in_bucket = GcpObject.from_api(obj)
            client.delete(
                object_in_bucket.api_spec.for_delete(),
                bucket=self.name,
                resource=object_in_bucket.name,
            )
        return True

    def delete(self, graph: Graph) -> bool:
        client = get_client(self)
        api_spec = self.api_spec.for_delete()
        api_spec.request_parameter = {"bucket": "{bucket}"}
        client.delete(
            api_spec,
            bucket=self.name,
        )
        return True

    def update_tag(self, key: str, value: Optional[str]) -> bool:
        client = get_client(self)

        labels = dict(self.tags)
        if value is None:
            if key in labels:
                del labels[key]
            else:
                return False
        else:
            labels.update({key: value})

        api_spec = self.api_spec.for_set_labels()
        api_spec.action = "patch"
        api_spec.request_parameter = {"bucket": "{bucket}"}
        client.set_labels(
            api_spec,
            body={"labels": labels},
            bucket=self.name,
        )
        return True

    def delete_tag(self, key: str) -> bool:
        return self.update_tag(key, None)


resources = [GcpBucket]



from datetime import datetime
from typing import ClassVar, Dict, Optional, List

from attr import define, field

from fix_plugin_gcp.gcp_client import GcpApiSpec
from fix_plugin_gcp.resources.base import GcpResource, GcpDeprecationStatus, GraphBuilder
from fixlib.baseresources import ModelReference, BaseManagedKubernetesClusterProvider
from fixlib.json_bender import Bender, S, Bend, ForallBend, MapDict
from fixlib.types import Json

# This service is called Google Kubernetes Engine in the docs
# https://cloud.google.com/kubernetes-engine/docs


@define(eq=False, slots=False)
class GcpContainerCloudRunConfig:
    kind: ClassVar[str] = "gcp_container_cloud_run_config"
    kind_display: ClassVar[str] = "GCP Container Cloud Run Configuration"
    kind_description: ClassVar[str] = (
        "GCP Container Cloud Run Config allows users to define and configure runtime"
        " settings for applications running on Google Cloud's serverless platform,"
        " Cloud Run."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"disabled": S("disabled"), "load_balancer_type": S("loadBalancerType")}
    disabled: Optional[bool] = field(default=None)
    load_balancer_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerAddonsConfig:
    kind: ClassVar[str] = "gcp_container_addons_config"
    kind_display: ClassVar[str] = "GCP Container Addons Config"
    kind_description: ClassVar[str] = (
        "GCP Container Addons Config is a configuration setting in Google Cloud"
        " Platform that allows users to enable or disable add-ons for Kubernetes"
        " Engine clusters."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "cloud_run_config": S("cloudRunConfig", default={}) >> Bend(GcpContainerCloudRunConfig.mapping),
        "config_connector_config": S("configConnectorConfig", "enabled"),
        "dns_cache_config": S("dnsCacheConfig", "enabled"),
        "gce_persistent_disk_csi_driver_config": S("gcePersistentDiskCsiDriverConfig", "enabled"),
        "gcp_filestore_csi_driver_config": S("gcpFilestoreCsiDriverConfig", "enabled"),
        "gke_backup_agent_config": S("gkeBackupAgentConfig", "enabled"),
        "horizontal_pod_autoscaling": S("horizontalPodAutoscaling", "disabled"),
        "http_load_balancing": S("httpLoadBalancing", "disabled"),
        "kubernetes_dashboard": S("kubernetesDashboard", "disabled"),
        "network_policy_config": S("networkPolicyConfig", "disabled"),
    }
    cloud_run_config: Optional[GcpContainerCloudRunConfig] = field(default=None)
    config_connector_config: Optional[bool] = field(default=None)
    dns_cache_config: Optional[bool] = field(default=None)
    gce_persistent_disk_csi_driver_config: Optional[bool] = field(default=None)
    gcp_filestore_csi_driver_config: Optional[bool] = field(default=None)
    gke_backup_agent_config: Optional[bool] = field(default=None)
    horizontal_pod_autoscaling: Optional[bool] = field(default=None)
    http_load_balancing: Optional[bool] = field(default=None)
    kubernetes_dashboard: Optional[bool] = field(default=None)
    network_policy_config: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerAuthenticatorGroupsConfig:
    kind: ClassVar[str] = "gcp_container_authenticator_groups_config"
    kind_display: ClassVar[str] = "GCP Container Authenticator Groups Config"
    kind_description: ClassVar[str] = (
        "In the context of GCP Container, the Authenticator Groups Config specifies whether"
        " a security group is enabled for authenticating containers."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("enabled"), "security_group": S("securityGroup")}
    enabled: Optional[bool] = field(default=None)
    security_group: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerAutoUpgradeOptions:
    kind: ClassVar[str] = "gcp_container_auto_upgrade_options"
    kind_display: ClassVar[str] = "GCP Container Auto-Upgrade Options"
    kind_description: ClassVar[str] = (
        "GCP Container Auto-Upgrade Options refer to the settings available for"
        " automatically upgrading Kubernetes clusters in the Google Cloud Platform,"
        " ensuring that they are always running the latest version of Kubernetes for"
        " enhanced security and performance."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "auto_upgrade_start_time": S("autoUpgradeStartTime"),
        "description": S("description"),
    }
    auto_upgrade_start_time: Optional[datetime] = field(default=None)
    description: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodeManagement:
    kind: ClassVar[str] = "gcp_container_node_management"
    kind_display: ClassVar[str] = "GCP Container Node Management"
    kind_description: ClassVar[str] = (
        "GCP Container Node Management is a service provided by Google Cloud Platform"
        " for managing and orchestrating containers running on GCP Kubernetes Engine."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "auto_repair": S("autoRepair"),
        "auto_upgrade": S("autoUpgrade"),
        "upgrade_options": S("upgradeOptions", default={}) >> Bend(GcpContainerAutoUpgradeOptions.mapping),
    }
    auto_repair: Optional[bool] = field(default=None)
    auto_upgrade: Optional[bool] = field(default=None)
    upgrade_options: Optional[GcpContainerAutoUpgradeOptions] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerShieldedInstanceConfig:
    kind: ClassVar[str] = "gcp_container_shielded_instance_config"
    kind_display: ClassVar[str] = "GCP Container Shielded Instance Config"
    kind_description: ClassVar[str] = (
        "The GCP Container Shielded Instance Config ensures enhanced security for GKE nodes by offering"
        " options to enable integrity monitoring and secure boot features."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "enable_integrity_monitoring": S("enableIntegrityMonitoring"),
        "enable_secure_boot": S("enableSecureBoot"),
    }
    enable_integrity_monitoring: Optional[bool] = field(default=None)
    enable_secure_boot: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerStandardRolloutPolicy:
    kind: ClassVar[str] = "gcp_container_standard_rollout_policy"
    kind_display: ClassVar[str] = "GCP Container Standard Rollout Policy"
    kind_description: ClassVar[str] = (
        "A rollout policy in Google Cloud Platform (GCP) Container is a standard"
        " mechanism that defines how new versions of a container should be gradually"
        " deployed to a cluster in a controlled manner."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "batch_node_count": S("batchNodeCount"),
        "batch_percentage": S("batchPercentage"),
        "batch_soak_duration": S("batchSoakDuration"),
    }
    batch_node_count: Optional[int] = field(default=None)
    batch_percentage: Optional[float] = field(default=None)
    batch_soak_duration: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerBlueGreenSettings:
    kind: ClassVar[str] = "gcp_container_blue_green_settings"
    kind_display: ClassVar[str] = "GCP Container Blue-Green Settings"
    kind_description: ClassVar[str] = (
        "GCP Container Blue-Green Settings refer to the configuration options for managing the node"
        " pool upgrade process, including the soak time for new nodes and the policy for rolling out"
        " upgrades across node pools."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "node_pool_soak_duration": S("nodePoolSoakDuration"),
        "standard_rollout_policy": S("standardRolloutPolicy", default={})
        >> Bend(GcpContainerStandardRolloutPolicy.mapping),
    }
    node_pool_soak_duration: Optional[str] = field(default=None)
    standard_rollout_policy: Optional[GcpContainerStandardRolloutPolicy] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerUpgradeSettings:
    kind: ClassVar[str] = "gcp_container_upgrade_settings"
    kind_display: ClassVar[str] = "GCP Container Upgrade Settings"
    kind_description: ClassVar[str] = (
        "GCP Container Upgrade Settings are configurations that allow users to manage"
        " and control the upgrade process of their containerized applications in"
        " Google Cloud Platform."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "blue_green_settings": S("blueGreenSettings", default={}) >> Bend(GcpContainerBlueGreenSettings.mapping),
        "max_surge": S("maxSurge"),
        "max_unavailable": S("maxUnavailable"),
        "strategy": S("strategy"),
    }
    blue_green_settings: Optional[GcpContainerBlueGreenSettings] = field(default=None)
    max_surge: Optional[int] = field(default=None)
    max_unavailable: Optional[int] = field(default=None)
    strategy: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerAutoprovisioningNodePoolDefaults:
    kind: ClassVar[str] = "gcp_container_autoprovisioning_node_pool_defaults"
    kind_display: ClassVar[str] = "GCP Container Autoprovisioning Node Pool Defaults"
    kind_description: ClassVar[str] = (
        "In GCP Container Autoprovisioning, Node Pool Defaults determine the configuration for nodes created"
        " automatically, including settings for disk encryption, disk size and type, image type, access scopes,"
        " and instance configuration options."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "boot_disk_kms_key": S("bootDiskKmsKey"),
        "disk_size_gb": S("diskSizeGb"),
        "disk_type": S("diskType"),
        "image_type": S("imageType"),
        "management": S("management", default={}) >> Bend(GcpContainerNodeManagement.mapping),
        "min_cpu_platform": S("minCpuPlatform"),
        "oauth_scopes": S("oauthScopes", default=[]),
        "service_account": S("serviceAccount"),
        "shielded_instance_config": S("shieldedInstanceConfig", default={})
        >> Bend(GcpContainerShieldedInstanceConfig.mapping),
        "upgrade_settings": S("upgradeSettings", default={}) >> Bend(GcpContainerUpgradeSettings.mapping),
    }
    boot_disk_kms_key: Optional[str] = field(default=None)
    disk_size_gb: Optional[int] = field(default=None)
    disk_type: Optional[str] = field(default=None)
    image_type: Optional[str] = field(default=None)
    management: Optional[GcpContainerNodeManagement] = field(default=None)
    min_cpu_platform: Optional[str] = field(default=None)
    oauth_scopes: Optional[List[str]] = field(default=None)
    service_account: Optional[str] = field(default=None)
    shielded_instance_config: Optional[GcpContainerShieldedInstanceConfig] = field(default=None)
    upgrade_settings: Optional[GcpContainerUpgradeSettings] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerResourceLimit:
    kind: ClassVar[str] = "gcp_container_resource_limit"
    kind_display: ClassVar[str] = "GCP Container Resource Limit"
    kind_description: ClassVar[str] = (
        "Container Resource Limit in Google Cloud Platform (GCP) is a feature that"
        " allows you to set resource constraints on containers, such as CPU and memory"
        " limits, to ensure efficient resource allocation and prevent resource"
        " starvation."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "maximum": S("maximum"),
        "minimum": S("minimum"),
        "resource_type": S("resourceType"),
    }
    maximum: Optional[str] = field(default=None)
    minimum: Optional[str] = field(default=None)
    resource_type: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerClusterAutoscaling:
    kind: ClassVar[str] = "gcp_container_cluster_autoscaling"
    kind_display: ClassVar[str] = "GCP Container Cluster Autoscaling"
    kind_description: ClassVar[str] = (
        "Container Cluster Autoscaling is a feature in Google Cloud Platform (GCP)"
        " that dynamically adjusts the number of nodes in a container cluster based on"
        " application demand and resource utilization."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "autoprovisioning_locations": S("autoprovisioningLocations", default=[]),
        "autoprovisioning_node_pool_defaults": S("autoprovisioningNodePoolDefaults", default={})
        >> Bend(GcpContainerAutoprovisioningNodePoolDefaults.mapping),
        "autoscaling_profile": S("autoscalingProfile"),
        "enable_node_autoprovisioning": S("enableNodeAutoprovisioning"),
        "resource_limits": S("resourceLimits", default=[]) >> ForallBend(GcpContainerResourceLimit.mapping),
    }
    autoprovisioning_locations: Optional[List[str]] = field(default=None)
    autoprovisioning_node_pool_defaults: Optional[GcpContainerAutoprovisioningNodePoolDefaults] = field(default=None)
    autoscaling_profile: Optional[str] = field(default=None)
    enable_node_autoprovisioning: Optional[bool] = field(default=None)
    resource_limits: Optional[List[GcpContainerResourceLimit]] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerBinaryAuthorization:
    kind: ClassVar[str] = "gcp_container_binary_authorization"
    kind_display: ClassVar[str] = "GCP Container Binary Authorization"
    kind_description: ClassVar[str] = (
        "GCP Container Binary Authorization is a service that ensures only trusted"
        " container images are deployed in your Google Cloud environment, helping to"
        " prevent unauthorized or vulnerable images from running in production."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("enabled"), "evaluation_mode": S("evaluationMode")}
    enabled: Optional[bool] = field(default=None)
    evaluation_mode: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerStatusCondition:
    kind: ClassVar[str] = "gcp_container_status_condition"
    kind_display: ClassVar[str] = "GCP Container Status Condition"
    kind_description: ClassVar[str] = (
        "The GCP Container Status Condition provides detailed status codes and messages that"
        " help to diagnose the health and operational state of GKE node pools and operations."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "canonical_code": S("canonicalCode"),
        "code": S("code"),
        "message": S("message"),
    }
    canonical_code: Optional[str] = field(default=None)
    code: Optional[str] = field(default=None)
    message: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerDatabaseEncryption:
    kind: ClassVar[str] = "gcp_container_database_encryption"
    kind_display: ClassVar[str] = "GCP Container Database Encryption"
    kind_description: ClassVar[str] = (
        "GCP Container Database Encryption settings in a container cluster define"
        " the encryption key used for database encryption and its operational state."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"key_name": S("keyName"), "state": S("state")}
    key_name: Optional[str] = field(default=None)
    state: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerIPAllocationPolicy:
    kind: ClassVar[str] = "gcp_container_ip_allocation_policy"
    kind_display: ClassVar[str] = "GCP Container IP Allocation Policy"
    kind_description: ClassVar[str] = (
        "Container IP Allocation Policy is a feature in Google Cloud Platform that"
        " allows users to define and manage the IP address allocation policy for"
        " containers in a Kubernetes cluster."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "cluster_ipv4_cidr": S("clusterIpv4Cidr"),
        "cluster_ipv4_cidr_block": S("clusterIpv4CidrBlock"),
        "cluster_secondary_range_name": S("clusterSecondaryRangeName"),
        "create_subnetwork": S("createSubnetwork"),
        "ipv6_access_type": S("ipv6AccessType"),
        "node_ipv4_cidr": S("nodeIpv4Cidr"),
        "node_ipv4_cidr_block": S("nodeIpv4CidrBlock"),
        "services_ipv4_cidr": S("servicesIpv4Cidr"),
        "services_ipv4_cidr_block": S("servicesIpv4CidrBlock"),
        "services_secondary_range_name": S("servicesSecondaryRangeName"),
        "stack_type": S("stackType"),
        "subnetwork_name": S("subnetworkName"),
        "tpu_ipv4_cidr_block": S("tpuIpv4CidrBlock"),
        "use_ip_aliases": S("useIpAliases"),
        "use_routes": S("useRoutes"),
    }
    cluster_ipv4_cidr: Optional[str] = field(default=None)
    cluster_ipv4_cidr_block: Optional[str] = field(default=None)
    cluster_secondary_range_name: Optional[str] = field(default=None)
    create_subnetwork: Optional[bool] = field(default=None)
    ipv6_access_type: Optional[str] = field(default=None)
    node_ipv4_cidr: Optional[str] = field(default=None)
    node_ipv4_cidr_block: Optional[str] = field(default=None)
    services_ipv4_cidr: Optional[str] = field(default=None)
    services_ipv4_cidr_block: Optional[str] = field(default=None)
    services_secondary_range_name: Optional[str] = field(default=None)
    stack_type: Optional[str] = field(default=None)
    subnetwork_name: Optional[str] = field(default=None)
    tpu_ipv4_cidr_block: Optional[str] = field(default=None)
    use_ip_aliases: Optional[bool] = field(default=None)
    use_routes: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerLoggingComponentConfig:
    kind: ClassVar[str] = "gcp_container_logging_component_config"
    kind_display: ClassVar[str] = "GCP Container Logging Component Config"
    kind_description: ClassVar[str] = (
        "Container Logging Component Config is a configuration setting for logging"
        " containers in the Google Cloud Platform."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"enable_components": S("enableComponents", default=[])}
    enable_components: Optional[List[str]] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerLoggingConfig:
    kind: ClassVar[str] = "gcp_container_logging_config"
    kind_display: ClassVar[str] = "GCP Container Logging Config"
    kind_description: ClassVar[str] = (
        "Container Logging Config is a feature in Google Cloud Platform (GCP) that"
        " allows users to configure and manage logging for their containerized"
        " applications running on GCP's Kubernetes Engine clusters."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "component_config": S("componentConfig", default={}) >> Bend(GcpContainerLoggingComponentConfig.mapping)
    }
    component_config: Optional[GcpContainerLoggingComponentConfig] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerDailyMaintenanceWindow:
    kind: ClassVar[str] = "gcp_container_daily_maintenance_window"
    kind_display: ClassVar[str] = "GCP Container Daily Maintenance Window"
    kind_description: ClassVar[str] = (
        "This resource represents the daily maintenance window for Google Cloud"
        " Platform (GCP) containers, during which routine maintenance activities can"
        " take place."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"duration": S("duration"), "start_time": S("startTime")}
    duration: Optional[str] = field(default=None)
    start_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerTimeWindow:
    kind: ClassVar[str] = "gcp_container_time_window"
    kind_display: ClassVar[str] = "GCP Container Time Window"
    kind_description: ClassVar[str] = (
        "The GCP Container Time Window specifies the start and end times for maintenance windows, including"
        " any specific exclusions to accommodate operational preferences within Google Kubernetes Engine."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "end_time": S("endTime"),
        "maintenance_exclusion_options": S("maintenanceExclusionOptions", "scope"),
        "start_time": S("startTime"),
    }
    end_time: Optional[datetime] = field(default=None)
    maintenance_exclusion_options: Optional[str] = field(default=None)
    start_time: Optional[datetime] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerRecurringTimeWindow:
    kind: ClassVar[str] = "gcp_container_recurring_time_window"
    kind_display: ClassVar[str] = "GCP Container Recurring Time Window"
    kind_description: ClassVar[str] = (
        "Container Recurring Time Window defines the schedule for regular maintenance operations."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "recurrence": S("recurrence"),
        "window": S("window", default={}) >> Bend(GcpContainerTimeWindow.mapping),
    }
    recurrence: Optional[str] = field(default=None)
    window: Optional[GcpContainerTimeWindow] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerMaintenanceWindow:
    kind: ClassVar[str] = "gcp_container_maintenance_window"
    kind_display: ClassVar[str] = "GCP Container Maintenance Window"
    kind_description: ClassVar[str] = (
        "The Container Maintenance Window specifies the preferred times and days for maintenance operations,"
        " including options for daily maintenance and recurring windows, along with any exclusions to the"
        " standard schedule."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "daily_maintenance_window": S("dailyMaintenanceWindow", default={})
        >> Bend(GcpContainerDailyMaintenanceWindow.mapping),
        "maintenance_exclusions": S("maintenanceExclusions", default={})
        >> MapDict(value_bender=Bend(GcpContainerTimeWindow.mapping)),
        "recurring_window": S("recurringWindow", default={}) >> Bend(GcpContainerRecurringTimeWindow.mapping),
    }
    daily_maintenance_window: Optional[GcpContainerDailyMaintenanceWindow] = field(default=None)
    maintenance_exclusions: Optional[Dict[str, GcpContainerTimeWindow]] = field(default=None)
    recurring_window: Optional[GcpContainerRecurringTimeWindow] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerMaintenancePolicy:
    kind: ClassVar[str] = "gcp_container_maintenance_policy"
    kind_display: ClassVar[str] = "GCP Container Maintenance Policy"
    kind_description: ClassVar[str] = (
        "GCP Container Maintenance Policy is a feature in Google Cloud Platform that"
        " allows users to define how their container clusters will be updated and"
        " maintained by specifying maintenance windows and auto-upgrade settings."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "resource_version": S("resourceVersion"),
        "window": S("window", default={}) >> Bend(GcpContainerMaintenanceWindow.mapping),
    }
    resource_version: Optional[str] = field(default=None)
    window: Optional[GcpContainerMaintenanceWindow] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerMasterAuth:
    kind: ClassVar[str] = "gcp_container_master_auth"
    kind_display: ClassVar[str] = "GCP Container Cluster Master Authentication"
    kind_description: ClassVar[str] = (
        "GCP Container Cluster Master Authentication provides secure access and"
        " authentication to the master controller of a Google Cloud Platform (GCP)"
        " container cluster, allowing users to manage and control their container"
        " cluster resources."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "client_certificate": S("clientCertificate"),
        "client_certificate_config": S("clientCertificateConfig", "issueClientCertificate"),
        "client_key": S("clientKey"),
        "cluster_ca_certificate": S("clusterCaCertificate"),
        "password": S("password"),
        "username": S("username"),
    }
    client_certificate: Optional[str] = field(default=None)
    client_certificate_config: Optional[bool] = field(default=None)
    client_key: Optional[str] = field(default=None)
    cluster_ca_certificate: Optional[str] = field(default=None)
    password: Optional[str] = field(default=None)
    username: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerCidrBlock:
    kind: ClassVar[str] = "gcp_container_cidr_block"
    kind_display: ClassVar[str] = "GCP Container CIDR Block"
    kind_description: ClassVar[str] = (
        "GCP Container CIDR Block is a range of IP addresses that can be used for the"
        " pods within a Google Cloud Platform (GCP) container cluster."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"cidr_block": S("cidrBlock"), "display_name": S("displayName")}
    cidr_block: Optional[str] = field(default=None)
    display_name: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerMasterAuthorizedNetworksConfig:
    kind: ClassVar[str] = "gcp_container_master_authorized_networks_config"
    kind_display: ClassVar[str] = "GCP Container Master Authorized Networks Configuration"
    kind_description: ClassVar[str] = (
        "Container Master Authorized Networks Configuration allows you to configure"
        " the IP address ranges that have access to the Kubernetes master of a Google"
        " Cloud Platform (GCP) container."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "cidr_blocks": S("cidrBlocks", default=[]) >> ForallBend(GcpContainerCidrBlock.mapping),
        "enabled": S("enabled"),
    }
    cidr_blocks: Optional[List[GcpContainerCidrBlock]] = field(default=None)
    enabled: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerMonitoringComponentConfig:
    kind: ClassVar[str] = "gcp_container_monitoring_component_config"
    kind_display: ClassVar[str] = "GCP Container Monitoring Component Config"
    kind_description: ClassVar[str] = (
        "GCP Container Monitoring Component Config is a configuration component used"
        " for monitoring containers in the Google Cloud Platform. It allows users to"
        " configure various settings and parameters for container monitoring."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"enable_components": S("enableComponents", default=[])}
    enable_components: Optional[List[str]] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerMonitoringConfig:
    kind: ClassVar[str] = "gcp_container_monitoring_config"
    kind_display: ClassVar[str] = "GCP Container Monitoring Config"
    kind_description: ClassVar[str] = (
        "GCP Container Monitoring Config is a feature provided by Google Cloud"
        " Platform that allows users to configure and monitor the containers running"
        " on their cloud infrastructure."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "component_config": S("componentConfig", default={}) >> Bend(GcpContainerMonitoringComponentConfig.mapping),
        "managed_prometheus_config": S("managedPrometheusConfig", "enabled"),
    }
    component_config: Optional[GcpContainerMonitoringComponentConfig] = field(default=None)
    managed_prometheus_config: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerDNSConfig:
    kind: ClassVar[str] = "gcp_container_dns_config"
    kind_display: ClassVar[str] = "GCP Container DNS Config"
    kind_description: ClassVar[str] = (
        "Container DNS Config is a feature in Google Cloud Platform that allows users"
        " to configure DNS settings for containers running in Google Kubernetes Engine"
        " (GKE)."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "cluster_dns": S("clusterDns"),
        "cluster_dns_domain": S("clusterDnsDomain"),
        "cluster_dns_scope": S("clusterDnsScope"),
    }
    cluster_dns: Optional[str] = field(default=None)
    cluster_dns_domain: Optional[str] = field(default=None)
    cluster_dns_scope: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNetworkConfig:
    kind: ClassVar[str] = "gcp_container_network_config"
    kind_display: ClassVar[str] = "GCP Container Network Config"
    kind_description: ClassVar[str] = (
        "Container Network Config is a feature provided by Google Cloud Platform that"
        " allows users to configure network settings for their containerized"
        " applications running in Google Kubernetes Engine (GKE), such as IP"
        " addresses, subnets, and network policies."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "datapath_provider": S("datapathProvider"),
        "default_snat_status": S("defaultSnatStatus", "disabled"),
        "dns_config": S("dnsConfig", default={}) >> Bend(GcpContainerDNSConfig.mapping),
        "enable_intra_node_visibility": S("enableIntraNodeVisibility"),
        "enable_l4ilb_subsetting": S("enableL4ilbSubsetting"),
        "network": S("network"),
        "private_ipv6_google_access": S("privateIpv6GoogleAccess"),
        "service_external_ips_config": S("serviceExternalIpsConfig", "enabled"),
        "subnetwork": S("subnetwork"),
    }
    datapath_provider: Optional[str] = field(default=None)
    default_snat_status: Optional[bool] = field(default=None)
    dns_config: Optional[GcpContainerDNSConfig] = field(default=None)
    enable_intra_node_visibility: Optional[bool] = field(default=None)
    enable_l4ilb_subsetting: Optional[bool] = field(default=None)
    network: Optional[str] = field(default=None)
    private_ipv6_google_access: Optional[str] = field(default=None)
    service_external_ips_config: Optional[bool] = field(default=None)
    subnetwork: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNetworkPolicy:
    kind: ClassVar[str] = "gcp_container_network_policy"
    kind_display: ClassVar[str] = "GCP Container Network Policy"
    kind_description: ClassVar[str] = (
        "GCP Container Network Policy is a resource in Google Cloud Platform that"
        " allows users to control network traffic between containers within a"
        " Kubernetes Engine cluster."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"enabled": S("enabled"), "provider": S("provider")}
    enabled: Optional[bool] = field(default=None)
    provider: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerGPUSharingConfig:
    kind: ClassVar[str] = "gcp_container_gpu_sharing_config"
    kind_display: ClassVar[str] = "GCP Container GPU Sharing Config"
    kind_description: ClassVar[str] = (
        "GCP Container GPU Sharing Config, within the scope of GCP Container Accelerator Config, determines how"
        " GPUs are shared among container instances and sets the maximum number of clients that can"
        " simultaneously utilize a single GPU."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "gpu_sharing_strategy": S("gpuSharingStrategy"),
        "max_shared_clients_per_gpu": S("maxSharedClientsPerGpu"),
    }
    gpu_sharing_strategy: Optional[str] = field(default=None)
    max_shared_clients_per_gpu: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerAcceleratorConfig:
    kind: ClassVar[str] = "gcp_container_accelerator_config"
    kind_display: ClassVar[str] = "GCP Container Accelerator Config"
    kind_description: ClassVar[str] = (
        "Container Accelerator Config is a feature in Google Cloud Platform that"
        " allows you to attach GPUs (Graphical Processing Units) to your containers,"
        " enabling faster and more efficient workload processing."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "accelerator_count": S("acceleratorCount"),
        "accelerator_type": S("acceleratorType"),
        "gpu_partition_size": S("gpuPartitionSize"),
        "gpu_sharing_config": S("gpuSharingConfig", default={}) >> Bend(GcpContainerGPUSharingConfig.mapping),
    }
    accelerator_count: Optional[str] = field(default=None)
    accelerator_type: Optional[str] = field(default=None)
    gpu_partition_size: Optional[str] = field(default=None)
    gpu_sharing_config: Optional[GcpContainerGPUSharingConfig] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodeKubeletConfig:
    kind: ClassVar[str] = "gcp_container_node_kubelet_config"
    kind_display: ClassVar[str] = "GCP Container Node Kubelet Config"
    kind_description: ClassVar[str] = (
        "The GCP Container Node Kubelet Config is a configuration file used by Google"
        " Cloud Platform (GCP) to configure the Kubelet component of container nodes"
        " in a Kubernetes cluster. Kubelet is responsible for managing the state of"
        " each container running on the node."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "cpu_cfs_quota": S("cpuCfsQuota"),
        "cpu_cfs_quota_period": S("cpuCfsQuotaPeriod"),
        "cpu_manager_policy": S("cpuManagerPolicy"),
        "pod_pids_limit": S("podPidsLimit"),
    }
    cpu_cfs_quota: Optional[bool] = field(default=None)
    cpu_cfs_quota_period: Optional[str] = field(default=None)
    cpu_manager_policy: Optional[str] = field(default=None)
    pod_pids_limit: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerLinuxNodeConfig:
    kind: ClassVar[str] = "gcp_container_linux_node_config"
    kind_display: ClassVar[str] = "GCP Container Linux Node Config"
    kind_description: ClassVar[str] = (
        "GCP Container Linux Node Config is a configuration for Linux nodes in Google"
        " Cloud Platform's container service, allowing users to define the settings"
        " and behavior for their Linux-based container nodes."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"sysctls": S("sysctls")}
    sysctls: Optional[Dict[str, str]] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodePoolLoggingConfig:
    kind: ClassVar[str] = "gcp_container_node_pool_logging_config"
    kind_display: ClassVar[str] = "GCP Container Node Pool Logging Config"
    kind_description: ClassVar[str] = (
        "Container Node Pool Logging Config is a configuration setting in Google"
        " Cloud Platform (GCP) for specifying logging options for container node pools"
        " in Kubernetes clusters."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"variant_config": S("variantConfig", "variant")}
    variant_config: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerReservationAffinity:
    kind: ClassVar[str] = "gcp_container_reservation_affinity"
    kind_display: ClassVar[str] = "GCP Container Reservation Affinity"
    kind_description: ClassVar[str] = (
        "Container Reservation Affinity is a setting that controls how instances are scheduled on reservations."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "consume_reservation_type": S("consumeReservationType"),
        "key": S("key"),
        "values": S("values", default=[]),
    }
    consume_reservation_type: Optional[str] = field(default=None)
    key: Optional[str] = field(default=None)
    values: Optional[List[str]] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodeTaint:
    kind: ClassVar[str] = "gcp_container_node_taint"
    kind_display: ClassVar[str] = "GCP Container Node Taint"
    kind_description: ClassVar[str] = (
        "Container Node Taints are a feature in Google Cloud Platform's container"
        " service that allow users to add constraints and preferences to nodes in a"
        " Kubernetes cluster."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"effect": S("effect"), "key": S("key"), "value": S("value")}
    effect: Optional[str] = field(default=None)
    key: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodeConfig:
    kind: ClassVar[str] = "gcp_container_node_config"
    kind_display: ClassVar[str] = "GCP Container Node Config"
    kind_description: ClassVar[str] = (
        "GCP Container Node Config is a configuration for a node in Google Cloud"
        " Platform's container service, allowing users to specify settings such as"
        " machine type, disk size, and network configuration for a container node."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "accelerators": S("accelerators", default=[]) >> ForallBend(GcpContainerAcceleratorConfig.mapping),
        "advanced_machine_features": S("advancedMachineFeatures", "threadsPerCore"),
        "boot_disk_kms_key": S("bootDiskKmsKey"),
        "confidential_nodes": S("confidentialNodes", "enabled"),
        "disk_size_gb": S("diskSizeGb"),
        "disk_type": S("diskType"),
        "gcfs_config": S("gcfsConfig", "enabled"),
        "gvnic": S("gvnic", "enabled"),
        "image_type": S("imageType"),
        "kubelet_config": S("kubeletConfig", default={}) >> Bend(GcpContainerNodeKubeletConfig.mapping),
        "labels": S("labels"),
        "linux_node_config": S("linuxNodeConfig", default={}) >> Bend(GcpContainerLinuxNodeConfig.mapping),
        "local_ssd_count": S("localSsdCount"),
        "logging_config": S("loggingConfig", default={}) >> Bend(GcpContainerNodePoolLoggingConfig.mapping),
        "machine_type": S("machineType"),
        "metadata": S("metadata"),
        "min_cpu_platform": S("minCpuPlatform"),
        "node_group": S("nodeGroup"),
        "oauth_scopes": S("oauthScopes", default=[]),
        "preemptible": S("preemptible"),
        "reservation_affinity": S("reservationAffinity", default={}) >> Bend(GcpContainerReservationAffinity.mapping),
        "sandbox_config": S("sandboxConfig", "type"),
        "service_account": S("serviceAccount"),
        "shielded_instance_config": S("shieldedInstanceConfig", default={})
        >> Bend(GcpContainerShieldedInstanceConfig.mapping),
        "spot": S("spot"),
        "tags": S("tags", default=[]),
        "taints": S("taints", default=[]) >> ForallBend(GcpContainerNodeTaint.mapping),
        "workload_metadata_config": S("workloadMetadataConfig", "mode"),
    }
    accelerators: Optional[List[GcpContainerAcceleratorConfig]] = field(default=None)
    advanced_machine_features: Optional[str] = field(default=None)
    boot_disk_kms_key: Optional[str] = field(default=None)
    confidential_nodes: Optional[bool] = field(default=None)
    disk_size_gb: Optional[int] = field(default=None)
    disk_type: Optional[str] = field(default=None)
    gcfs_config: Optional[bool] = field(default=None)
    gvnic: Optional[bool] = field(default=None)
    image_type: Optional[str] = field(default=None)
    kubelet_config: Optional[GcpContainerNodeKubeletConfig] = field(default=None)
    labels: Optional[Dict[str, str]] = field(default=None)
    linux_node_config: Optional[GcpContainerLinuxNodeConfig] = field(default=None)
    local_ssd_count: Optional[int] = field(default=None)
    logging_config: Optional[GcpContainerNodePoolLoggingConfig] = field(default=None)
    machine_type: Optional[str] = field(default=None)
    metadata: Optional[Dict[str, str]] = field(default=None)
    min_cpu_platform: Optional[str] = field(default=None)
    node_group: Optional[str] = field(default=None)
    oauth_scopes: Optional[List[str]] = field(default=None)
    preemptible: Optional[bool] = field(default=None)
    reservation_affinity: Optional[GcpContainerReservationAffinity] = field(default=None)
    sandbox_config: Optional[str] = field(default=None)
    service_account: Optional[str] = field(default=None)
    shielded_instance_config: Optional[GcpContainerShieldedInstanceConfig] = field(default=None)
    spot: Optional[bool] = field(default=None)
    tags: Optional[List[str]] = field(default=None)
    taints: Optional[List[GcpContainerNodeTaint]] = field(default=None)
    workload_metadata_config: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNetworkTags:
    kind: ClassVar[str] = "gcp_container_network_tags"
    kind_display: ClassVar[str] = "GCP Container Network Tags"
    kind_description: ClassVar[str] = (
        "Within the GCP Container Node Pool Auto Configuration, Container Network Tags are labels applied"
        " to virtual machine instances, allowing for the setup of network firewall rules and routes that"
        " apply to tagged instances within the container cluster."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"tags": S("tags", default=[])}
    tags: Optional[List[str]] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodePoolAutoConfig:
    kind: ClassVar[str] = "gcp_container_node_pool_auto_config"
    kind_display: ClassVar[str] = "GCP Container Node Pool Auto Config"
    kind_description: ClassVar[str] = (
        "The Container Node Pool Auto Config refers to settings that automatically determine the network tags"
        " for nodes within a node pool. These tags are used to configure network firewall rules that apply to"
        " the instances in the node pool."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "network_tags": S("networkTags", default={}) >> Bend(GcpContainerNetworkTags.mapping)
    }
    network_tags: Optional[GcpContainerNetworkTags] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodeConfigDefaults:
    kind: ClassVar[str] = "gcp_container_node_config_defaults"
    kind_display: ClassVar[str] = "GCP Container Node Config Defaults"
    kind_description: ClassVar[str] = (
        "GCP Container Node Config Defaults represents the default configuration"
        " settings for nodes in a Google Cloud Platform container cluster."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "gcfs_config": S("gcfsConfig", "enabled"),
        "logging_config": S("loggingConfig", default={}) >> Bend(GcpContainerNodePoolLoggingConfig.mapping),
    }
    gcfs_config: Optional[bool] = field(default=None)
    logging_config: Optional[GcpContainerNodePoolLoggingConfig] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodePoolDefaults:
    kind: ClassVar[str] = "gcp_container_node_pool_defaults"
    kind_display: ClassVar[str] = "GCP Container Node Pool Defaults"
    kind_description: ClassVar[str] = (
        "GCP Container Node Pool Defaults is a feature in Google Cloud Platform that"
        " allows users to set default configurations for their container node pools,"
        " which are groups of nodes that host containerized applications."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "node_config_defaults": S("nodeConfigDefaults", default={}) >> Bend(GcpContainerNodeConfigDefaults.mapping)
    }
    node_config_defaults: Optional[GcpContainerNodeConfigDefaults] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodePoolAutoscaling:
    kind: ClassVar[str] = "gcp_container_node_pool_autoscaling"
    kind_display: ClassVar[str] = "GCP Container Node Pool Autoscaling"
    kind_description: ClassVar[str] = (
        "Container Node Pool Autoscaling is a feature in Google Cloud Platform that"
        " automatically adjusts the number of nodes in a container cluster based on"
        " demand, ensuring optimal resource utilization and scalability."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "autoprovisioned": S("autoprovisioned"),
        "enabled": S("enabled"),
        "location_policy": S("locationPolicy"),
        "max_node_count": S("maxNodeCount"),
        "min_node_count": S("minNodeCount"),
        "total_max_node_count": S("totalMaxNodeCount"),
        "total_min_node_count": S("totalMinNodeCount"),
    }
    autoprovisioned: Optional[bool] = field(default=None)
    enabled: Optional[bool] = field(default=None)
    location_policy: Optional[str] = field(default=None)
    max_node_count: Optional[int] = field(default=None)
    min_node_count: Optional[int] = field(default=None)
    total_max_node_count: Optional[int] = field(default=None)
    total_min_node_count: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodeNetworkConfig:
    kind: ClassVar[str] = "gcp_container_node_network_config"
    kind_display: ClassVar[str] = "GCP Container Node Network Config"
    kind_description: ClassVar[str] = (
        "GCP Container Node Network Config is a network configuration for nodes in"
        " Google Cloud Platform's container service. It defines the network settings"
        " for containers running on the nodes."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "create_pod_range": S("createPodRange"),
        "network_performance_config": S("networkPerformanceConfig", "totalEgressBandwidthTier"),
        "pod_ipv4_cidr_block": S("podIpv4CidrBlock"),
        "pod_range": S("podRange"),
    }
    create_pod_range: Optional[bool] = field(default=None)
    network_performance_config: Optional[str] = field(default=None)
    pod_ipv4_cidr_block: Optional[str] = field(default=None)
    pod_range: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerBlueGreenInfo:
    kind: ClassVar[str] = "gcp_container_blue_green_info"
    kind_display: ClassVar[str] = "GCP Container Blue-Green Info"
    kind_description: ClassVar[str] = (
        "Blue-Green deployment strategy in Google Cloud Platform (GCP) container"
        " where two identical production environments, blue and green, are used to"
        " minimize downtime during software releases."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "blue_instance_group_urls": S("blueInstanceGroupUrls", default=[]),
        "blue_pool_deletion_start_time": S("bluePoolDeletionStartTime"),
        "green_instance_group_urls": S("greenInstanceGroupUrls", default=[]),
        "green_pool_version": S("greenPoolVersion"),
        "phase": S("phase"),
    }
    blue_instance_group_urls: Optional[List[str]] = field(default=None)
    blue_pool_deletion_start_time: Optional[datetime] = field(default=None)
    green_instance_group_urls: Optional[List[str]] = field(default=None)
    green_pool_version: Optional[str] = field(default=None)
    phase: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerUpdateInfo:
    kind: ClassVar[str] = "gcp_container_update_info"
    kind_display: ClassVar[str] = "GCP Container Update Info"
    kind_description: ClassVar[str] = (
        "GCP Container Update Info outlines details about blue-green deployment strategies"
        " for node pools in Google Kubernetes Engine."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "blue_green_info": S("blueGreenInfo", default={}) >> Bend(GcpContainerBlueGreenInfo.mapping)
    }
    blue_green_info: Optional[GcpContainerBlueGreenInfo] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNodePool:
    kind: ClassVar[str] = "gcp_container_node_pool"
    kind_display: ClassVar[str] = "GCP Container Node Pool"
    kind_description: ClassVar[str] = (
        "Container Node Pool is a resource in Google Cloud Platform that allows you"
        " to create and manage a pool of virtual machines to run your containerized"
        " applications in Google Kubernetes Engine."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "autoscaling": S("autoscaling", default={}) >> Bend(GcpContainerNodePoolAutoscaling.mapping),
        "conditions": S("conditions", default=[]) >> ForallBend(GcpContainerStatusCondition.mapping),
        "config": S("config", default={}) >> Bend(GcpContainerNodeConfig.mapping),
        "initial_node_count": S("initialNodeCount"),
        "instance_group_urls": S("instanceGroupUrls", default=[]),
        "locations": S("locations", default=[]),
        "management": S("management", default={}) >> Bend(GcpContainerNodeManagement.mapping),
        "max_pods_constraint": S("maxPodsConstraint", "maxPodsPerNode"),
        "name": S("name"),
        "network_config": S("networkConfig", default={}) >> Bend(GcpContainerNodeNetworkConfig.mapping),
        "pod_ipv4_cidr_size": S("podIpv4CidrSize"),
        "self_link": S("selfLink"),
        "status": S("status"),
        "status_message": S("statusMessage"),
        "update_info": S("updateInfo", default={}) >> Bend(GcpContainerUpdateInfo.mapping),
        "upgrade_settings": S("upgradeSettings", default={}) >> Bend(GcpContainerUpgradeSettings.mapping),
        "version": S("version"),
    }
    autoscaling: Optional[GcpContainerNodePoolAutoscaling] = field(default=None)
    conditions: Optional[List[GcpContainerStatusCondition]] = field(default=None)
    config: Optional[GcpContainerNodeConfig] = field(default=None)
    initial_node_count: Optional[int] = field(default=None)
    instance_group_urls: Optional[List[str]] = field(default=None)
    locations: Optional[List[str]] = field(default=None)
    management: Optional[GcpContainerNodeManagement] = field(default=None)
    max_pods_constraint: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)
    network_config: Optional[GcpContainerNodeNetworkConfig] = field(default=None)
    pod_ipv4_cidr_size: Optional[int] = field(default=None)
    self_link: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)
    status_message: Optional[str] = field(default=None)
    update_info: Optional[GcpContainerUpdateInfo] = field(default=None)
    upgrade_settings: Optional[GcpContainerUpgradeSettings] = field(default=None)
    version: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerFilter:
    kind: ClassVar[str] = "gcp_container_filter"
    kind_display: ClassVar[str] = "GCP Container Filter"
    kind_description: ClassVar[str] = (
        "GCP Container Filter specifies the type of events that should trigger the container,"
        " typically based on Pub/Sub messages."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"event_type": S("eventType", default=[])}
    event_type: Optional[List[str]] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerPubSub:
    kind: ClassVar[str] = "gcp_container_pub_sub"
    kind_display: ClassVar[str] = "GCP Container Pub/Sub"
    kind_description: ClassVar[str] = (
        "In the context of GCP Container Notification Config, the Container Pub/Sub setting enables you to configure"
        " Google Cloud's Pub/Sub for receiving notifications. When enabled, it allows you to specify a filter"
        " to determine the types of cluster events that should trigger notifications, and the Pub/Sub topic"
        " to which these notifications will be sent."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "enabled": S("enabled"),
        "filter": S("filter", default={}) >> Bend(GcpContainerFilter.mapping),
        "topic": S("topic"),
    }
    enabled: Optional[bool] = field(default=None)
    filter: Optional[GcpContainerFilter] = field(default=None)
    topic: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerNotificationConfig:
    kind: ClassVar[str] = "gcp_container_notification_config"
    kind_display: ClassVar[str] = "GCP Container Notification Config"
    kind_description: ClassVar[str] = (
        "The Container Notification Config is a setting that specifies the Pub/Sub topic to which notifications"
        " for the cluster are published."
    )
    mapping: ClassVar[Dict[str, Bender]] = {"pubsub": S("pubsub", default={}) >> Bend(GcpContainerPubSub.mapping)}
    pubsub: Optional[GcpContainerPubSub] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerPrivateClusterConfig:
    kind: ClassVar[str] = "gcp_container_private_cluster_config"
    kind_display: ClassVar[str] = "GCP Container Private Cluster Config"
    kind_description: ClassVar[str] = (
        "Private cluster configuration option for running Kubernetes clusters in"
        " Google Cloud Platform (GCP) container engine. Private clusters offer"
        " enhanced security by isolating the cluster's control plane and worker nodes"
        " from the public internet."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "enable_private_endpoint": S("enablePrivateEndpoint"),
        "enable_private_nodes": S("enablePrivateNodes"),
        "master_global_access_config": S("masterGlobalAccessConfig", "enabled"),
        "master_ipv4_cidr_block": S("masterIpv4CidrBlock"),
        "peering_name": S("peeringName"),
        "private_endpoint": S("privateEndpoint"),
        "public_endpoint": S("publicEndpoint"),
    }
    enable_private_endpoint: Optional[bool] = field(default=None)
    enable_private_nodes: Optional[bool] = field(default=None)
    master_global_access_config: Optional[bool] = field(default=None)
    master_ipv4_cidr_block: Optional[str] = field(default=None)
    peering_name: Optional[str] = field(default=None)
    private_endpoint: Optional[str] = field(default=None)
    public_endpoint: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerResourceUsageExportConfig:
    kind: ClassVar[str] = "gcp_container_resource_usage_export_config"
    kind_display: ClassVar[str] = "GCP Container Resource Usage Export Config"
    kind_description: ClassVar[str] = (
        "The GCP Container Resource Usage Export Config is a feature that facilitates the analysis of cluster"
        " resource usage by exporting data to BigQuery and providing options for detailed consumption tracking"
        " and network egress monitoring."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "bigquery_destination": S("bigqueryDestination", "datasetId"),
        "consumption_metering_config": S("consumptionMeteringConfig", "enabled"),
        "enable_network_egress_metering": S("enableNetworkEgressMetering"),
    }
    bigquery_destination: Optional[str] = field(default=None)
    consumption_metering_config: Optional[bool] = field(default=None)
    enable_network_egress_metering: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerCluster(BaseManagedKubernetesClusterProvider, GcpResource):
    kind: ClassVar[str] = "gcp_container_cluster"
    kind_display: ClassVar[str] = "GCP Container Cluster"
    kind_description: ClassVar[str] = (
        "Container Cluster is a managed Kubernetes cluster service provided by Google"
        " Cloud Platform, which allows users to deploy, manage, and scale"
        " containerized applications using Kubernetes."
    )
    api_spec: ClassVar[GcpApiSpec] = GcpApiSpec(
        service="container",
        version="v1",
        accessors=["projects", "locations", "clusters"],
        action="list",
        request_parameter={"parent": "projects/{project}/locations/-"},
        request_parameter_in={"project"},
        response_path="clusters",
        response_regional_sub_path=None,
        required_iam_permissions=["container.clusters.list"],
        mutate_iam_permissions=["container.clusters.update", "container.clusters.delete"],
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("name").or_else(S("id")).or_else(S("selfLink")),
        "tags": S("labels", default={}),
        "name": S("name"),
        "ctime": S("creationTimestamp"),
        "description": S("description"),
        "link": S("selfLink"),
        "label_fingerprint": S("labelFingerprint"),
        "deprecation_status": S("deprecated", default={}) >> Bend(GcpDeprecationStatus.mapping),
        "addons_config": S("addonsConfig", default={}) >> Bend(GcpContainerAddonsConfig.mapping),
        "authenticator_groups_config": S("authenticatorGroupsConfig", default={})
        >> Bend(GcpContainerAuthenticatorGroupsConfig.mapping),
        "autopilot": S("autopilot", "enabled"),
        "autoscaling": S("autoscaling", default={}) >> Bend(GcpContainerClusterAutoscaling.mapping),
        "binary_authorization": S("binaryAuthorization", default={}) >> Bend(GcpContainerBinaryAuthorization.mapping),
        "cluster_ipv4_cidr": S("clusterIpv4Cidr"),
        "conditions": S("conditions", default=[]) >> ForallBend(GcpContainerStatusCondition.mapping),
        "confidential_nodes": S("confidentialNodes", "enabled"),
        "cost_management_config": S("costManagementConfig", "enabled"),
        "create_time": S("createTime"),
        "current_master_version": S("currentMasterVersion"),
        "version": S("currentMasterVersion"),
        "current_node_count": S("currentNodeCount"),
        "current_node_version": S("currentNodeVersion"),
        "database_encryption": S("databaseEncryption", default={}) >> Bend(GcpContainerDatabaseEncryption.mapping),
        "default_max_pods_constraint": S("defaultMaxPodsConstraint", "maxPodsPerNode"),
        "enable_kubernetes_alpha": S("enableKubernetesAlpha"),
        "enable_tpu": S("enableTpu"),
        "endpoint": S("endpoint"),
        "etag": S("etag"),
        "expire_time": S("expireTime"),
        "identity_service_config": S("identityServiceConfig", "enabled"),
        "initial_cluster_version": S("initialClusterVersion"),
        "initial_node_count": S("initialNodeCount"),
        "instance_group_urls": S("instanceGroupUrls", default=[]),
        "ip_allocation_policy": S("ipAllocationPolicy", default={}) >> Bend(GcpContainerIPAllocationPolicy.mapping),
        "legacy_abac": S("legacyAbac", "enabled"),
        "location": S("location"),
        "locations": S("locations", default=[]),
        "logging_config": S("loggingConfig", default={}) >> Bend(GcpContainerLoggingConfig.mapping),
        "logging_service": S("loggingService"),
        "container_cluster_maintenance_policy": S("maintenancePolicy", default={})
        >> Bend(GcpContainerMaintenancePolicy.mapping),
        "master_auth": S("masterAuth", default={}) >> Bend(GcpContainerMasterAuth.mapping),
        "master_authorized_networks_config": S("masterAuthorizedNetworksConfig", default={})
        >> Bend(GcpContainerMasterAuthorizedNetworksConfig.mapping),
        "mesh_certificates": S("meshCertificates", "enableCertificates"),
        "monitoring_config": S("monitoringConfig", default={}) >> Bend(GcpContainerMonitoringConfig.mapping),
        "monitoring_service": S("monitoringService"),
        "network": S("network"),
        "network_config": S("networkConfig", default={}) >> Bend(GcpContainerNetworkConfig.mapping),
        "network_policy": S("networkPolicy", default={}) >> Bend(GcpContainerNetworkPolicy.mapping),
        "node_config": S("nodeConfig", default={}) >> Bend(GcpContainerNodeConfig.mapping),
        "node_ipv4_cidr_size": S("nodeIpv4CidrSize"),
        "node_pool_auto_config": S("nodePoolAutoConfig", default={}) >> Bend(GcpContainerNodePoolAutoConfig.mapping),
        "node_pool_defaults": S("nodePoolDefaults", default={}) >> Bend(GcpContainerNodePoolDefaults.mapping),
        "node_pools": S("nodePools", default=[]) >> ForallBend(GcpContainerNodePool.mapping),
        "notification_config": S("notificationConfig", default={}) >> Bend(GcpContainerNotificationConfig.mapping),
        "private_cluster_config": S("privateClusterConfig", default={})
        >> Bend(GcpContainerPrivateClusterConfig.mapping),
        "release_channel": S("releaseChannel", "channel"),
        "resource_labels": S("resourceLabels"),
        "resource_usage_export_config": S("resourceUsageExportConfig", default={})
        >> Bend(GcpContainerResourceUsageExportConfig.mapping),
        "services_ipv4_cidr": S("servicesIpv4Cidr"),
        "shielded_nodes": S("shieldedNodes", "enabled"),
        "status": S("status"),
        "status_message": S("statusMessage"),
        "subnetwork": S("subnetwork"),
        "tpu_ipv4_cidr_block": S("tpuIpv4CidrBlock"),
        "vertical_pod_autoscaling": S("verticalPodAutoscaling", "enabled"),
        "workload_identity_config": S("workloadIdentityConfig", "workloadPool"),
    }
    addons_config: Optional[GcpContainerAddonsConfig] = field(default=None)
    authenticator_groups_config: Optional[GcpContainerAuthenticatorGroupsConfig] = field(default=None)
    autopilot: Optional[bool] = field(default=None)
    autoscaling: Optional[GcpContainerClusterAutoscaling] = field(default=None)
    binary_authorization: Optional[GcpContainerBinaryAuthorization] = field(default=None)
    cluster_ipv4_cidr: Optional[str] = field(default=None)
    conditions: Optional[List[GcpContainerStatusCondition]] = field(default=None)
    confidential_nodes: Optional[bool] = field(default=None)
    cost_management_config: Optional[bool] = field(default=None)
    create_time: Optional[datetime] = field(default=None)
    current_master_version: Optional[str] = field(default=None)
    current_node_count: Optional[int] = field(default=None)
    current_node_version: Optional[str] = field(default=None)
    database_encryption: Optional[GcpContainerDatabaseEncryption] = field(default=None)
    default_max_pods_constraint: Optional[str] = field(default=None)
    enable_kubernetes_alpha: Optional[bool] = field(default=None)
    enable_tpu: Optional[bool] = field(default=None)
    etag: Optional[str] = field(default=None)
    expire_time: Optional[datetime] = field(default=None)
    identity_service_config: Optional[bool] = field(default=None)
    initial_cluster_version: Optional[str] = field(default=None)
    initial_node_count: Optional[int] = field(default=None)
    instance_group_urls: Optional[List[str]] = field(default=None)
    ip_allocation_policy: Optional[GcpContainerIPAllocationPolicy] = field(default=None)
    legacy_abac: Optional[bool] = field(default=None)
    location: Optional[str] = field(default=None)
    locations: Optional[List[str]] = field(default=None)
    logging_config: Optional[GcpContainerLoggingConfig] = field(default=None)
    logging_service: Optional[str] = field(default=None)
    container_cluster_maintenance_policy: Optional[GcpContainerMaintenancePolicy] = field(default=None)
    master_auth: Optional[GcpContainerMasterAuth] = field(default=None)
    master_authorized_networks_config: Optional[GcpContainerMasterAuthorizedNetworksConfig] = field(default=None)
    mesh_certificates: Optional[bool] = field(default=None)
    monitoring_config: Optional[GcpContainerMonitoringConfig] = field(default=None)
    monitoring_service: Optional[str] = field(default=None)
    network: Optional[str] = field(default=None)
    network_config: Optional[GcpContainerNetworkConfig] = field(default=None)
    network_policy: Optional[GcpContainerNetworkPolicy] = field(default=None)
    node_config: Optional[GcpContainerNodeConfig] = field(default=None)
    node_ipv4_cidr_size: Optional[int] = field(default=None)
    node_pool_auto_config: Optional[GcpContainerNodePoolAutoConfig] = field(default=None)
    node_pool_defaults: Optional[GcpContainerNodePoolDefaults] = field(default=None)
    node_pools: Optional[List[GcpContainerNodePool]] = field(default=None)
    notification_config: Optional[GcpContainerNotificationConfig] = field(default=None)
    private_cluster_config: Optional[GcpContainerPrivateClusterConfig] = field(default=None)
    release_channel: Optional[str] = field(default=None)
    resource_labels: Optional[Dict[str, str]] = field(default=None)
    resource_usage_export_config: Optional[GcpContainerResourceUsageExportConfig] = field(default=None)
    services_ipv4_cidr: Optional[str] = field(default=None)
    shielded_nodes: Optional[bool] = field(default=None)
    status: Optional[str] = field(default=None)
    status_message: Optional[str] = field(default=None)
    subnetwork: Optional[str] = field(default=None)
    tpu_ipv4_cidr_block: Optional[str] = field(default=None)
    vertical_pod_autoscaling: Optional[bool] = field(default=None)
    workload_identity_config: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerStatus:
    kind: ClassVar[str] = "gcp_container_status"
    kind_display: ClassVar[str] = "GCP Container Status"
    kind_description: ClassVar[str] = (
        "GCP Container Status provides information about the current status, health,"
        " and availability of containers running on Google Cloud Platform (GCP)."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "code": S("code"),
        "details": S("details", default=[]),
        "message": S("message"),
    }
    code: Optional[int] = field(default=None)
    details: Optional[List[Json]] = field(default=None)
    message: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerMetric:
    kind: ClassVar[str] = "gcp_container_metric"
    kind_display: ClassVar[str] = "GCP Container Metric"
    kind_description: ClassVar[str] = (
        "Container Metrics in Google Cloud Platform (GCP) are measurements of"
        " resource utilization and performance for containers running on GCP's managed"
        " Kubernetes Engine."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "double_value": S("doubleValue"),
        "int_value": S("intValue"),
        "name": S("name"),
        "string_value": S("stringValue"),
    }
    double_value: Optional[float] = field(default=None)
    int_value: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)
    string_value: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerOperationProgress:
    kind: ClassVar[str] = "gcp_container_operation_progress"
    kind_display: ClassVar[str] = "GCP Container Operation Progress"
    kind_description: ClassVar[str] = (
        "GCP Container Operation Progress refers to the status and progress of an"
        " operation involving containers in Google Cloud Platform. It provides"
        " information on the current state and completion progress of container-"
        " related operations."
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "metrics": S("metrics", default=[]) >> ForallBend(GcpContainerMetric.mapping),
        "name": S("name"),
        "status": S("status"),
    }
    metrics: Optional[List[GcpContainerMetric]] = field(default=None)
    name: Optional[str] = field(default=None)
    status: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class GcpContainerOperation(GcpResource):
    kind: ClassVar[str] = "gcp_container_operation"
    kind_display: ClassVar[str] = "GCP Container Operation"
    kind_description: ClassVar[str] = (
        "Container Operations are management tasks performed on containers in Google"
        " Cloud Platform, including creating, starting, stopping, and deleting"
        " containers."
    )
    reference_kinds: ClassVar[ModelReference] = {"predecessors": {"default": ["gcp_container_cluster"]}}
    api_spec: ClassVar[GcpApiSpec] = GcpApiSpec(
        service="container",
        version="v1",
        accessors=["projects", "locations", "operations"],
        action="list",
        request_parameter={"parent": "projects/{project}/locations/-"},
        request_parameter_in={"project"},
        response_path="operations",
        response_regional_sub_path=None,
        required_iam_permissions=["container.operations.list"],
        mutate_iam_permissions=[],
    )
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("name").or_else(S("id")).or_else(S("selfLink")),
        "tags": S("labels", default={}),
        "name": S("name"),
        "ctime": S("creationTimestamp"),
        "description": S("description"),
        "link": S("selfLink"),
        "label_fingerprint": S("labelFingerprint"),
        "deprecation_status": S("deprecated", default={}) >> Bend(GcpDeprecationStatus.mapping),
        "cluster_conditions": S("clusterConditions", default=[]) >> ForallBend(GcpContainerStatusCondition.mapping),
        "detail": S("detail"),
        "end_time": S("endTime"),
        "container_operation_error": S("error", default={}) >> Bend(GcpContainerStatus.mapping),
        "location": S("location"),
        "nodepool_conditions": S("nodepoolConditions", default=[]) >> ForallBend(GcpContainerStatusCondition.mapping),
        "operation_type": S("operationType"),
        "container_operation_progress": S("progress", default={}) >> Bend(GcpContainerOperationProgress.mapping),
        "start_time": S("startTime"),
        "status": S("status"),
        "status_message": S("statusMessage"),
        "target_link": S("targetLink"),
    }
    cluster_conditions: Optional[List[GcpContainerStatusCondition]] = field(default=None)
    detail: Optional[str] = field(default=None)
    end_time: Optional[datetime] = field(default=None)
    container_operation_error: Optional[GcpContainerStatus] = field(default=None)
    location: Optional[str] = field(default=None)
    nodepool_conditions: Optional[List[GcpContainerStatusCondition]] = field(default=None)
    operation_type: Optional[str] = field(default=None)
    container_operation_progress: Optional[GcpContainerOperationProgress] = field(default=None)
    start_time: Optional[datetime] = field(default=None)
    status: Optional[str] = field(default=None)
    status_message: Optional[str] = field(default=None)
    target_link: Optional[str] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if self.target_link:
            builder.add_edge(self, reverse=True, clazz=GcpContainerCluster, link=self.target_link)


resources = [GcpContainerCluster, GcpContainerOperation]



from .random_client import connect_resource, roundtrip
from fix_plugin_gcp.resources.container import *
from fix_plugin_gcp.resources.base import GraphBuilder


def test_gcp_container_cluster(random_builder: GraphBuilder) -> None:
    roundtrip(GcpContainerCluster, random_builder)


def test_gcp_container_operation(random_builder: GraphBuilder) -> None:
    op = roundtrip(GcpContainerOperation, random_builder)
    connect_resource(random_builder, op, GcpContainerCluster, selfLink=op.target_link)
    assert len(random_builder.edges_of(GcpContainerCluster, GcpContainerOperation)) == 1
